# distutils: language=c++
# cython: boundscheck=False
# cython: cdivision=True
# cython: nonecheck=False
# cython: wraparound=False
# cython: language_level=3


## TODO: we are not exposing these functions (at least, officially),
## TODO: hence, they are not included in the online manual.
## TODO: Otherwise, the docstrings should be cleaned up
## TODO: and formatted so as to conform to the numpydoc guidelines.
## TODO: (volunteers needed) Cheers.



"""
Internal functions
"""

# Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License
# Version 3, 19 November 2007, published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License Version 3 for more details.
# You should have received a copy of the License along with this program.
# If not, see <https://www.gnu.org/licenses/>.






cimport cython
cimport numpy as np
import numpy as np


cimport libc.math
from libcpp cimport bool
from libcpp.vector cimport vector
from numpy.math cimport INFINITY


ctypedef fused T:
    int
    long
    long long
    ssize_t
    float
    double

ctypedef fused floatT:
    float
    double




from . cimport c_mst
from . cimport c_preprocess
from . cimport c_postprocess
from . cimport c_disjoint_sets
from . cimport c_gini_disjoint_sets
from . cimport c_genie

cimport openmp
import os

################################################################################
# Minimum Spanning Tree Algorithms:
# (a) Prim-Jarník's for Complete Undirected Graphs,
# (b) Kruskal's for k-NN graphs,
# and auxiliary functions.
################################################################################

cdef void _openmp_set_num_threads():
    cdef int nt = int(os.getenv('OMP_NUM_THREADS',
                                openmp.omp_get_max_threads()))
    openmp.omp_set_num_threads(nt)


cpdef np.ndarray[floatT] get_d_core(
    floatT[:,::1] dist,
    ssize_t[:,::1] ind,
    ssize_t M):
    """
    Get "core" distance = distance to the (M-1)-th nearest neighbour
    (if available, otherwise, distance to the furthest away one at hand).


    Parameters
    ----------

    dist : a c_contiguous ndarray, shape (n,k)
        dist[i,:] is sorted nondecreasingly for all i,
        dist[i,j] gives the weight of the edge {i, ind[i,j]}
    ind : a c_contiguous ndarray, shape (n,k)
        edge definition, interpreted as {i, ind[i,j]};
        -1 denotes a "missing value"
    M : int
        "smoothing factor"


    Returns
    -------

    ndarray
        of length dist.shape[0]
    """

    cdef ssize_t n = dist.shape[0]
    cdef ssize_t k = dist.shape[1]

    if not (ind.shape[0] == n and ind.shape[1] == k):
        raise ValueError("shapes of dist and ind must match")

    if M-2 >= k:
        raise ValueError("too few nearest neighbours provided")

    cdef np.ndarray[floatT] d_core = np.empty(n,
        dtype=np.float32 if floatT is float else np.float64)

    #Python equivalent if all NNs are available:
    #assert nn_dist.shape[1] >= cur_state["M"]-1
    #d_core = nn_dist[:, cur_state["M"]-2].astype(X.dtype, order="C")

    cdef ssize_t i, j
    for i in range(n):
        j = M-2
        while ind[i, j] < 0:
            j -= 1
            if j < 0: raise ValueError("no nearest neighbours provided")
        d_core[i] = dist[i, j]

    return d_core



cpdef tuple nn_list_to_matrix(
    list nns,
    ssize_t k_max):
    """
    Converts a list of (<=k_max)-nearest neighbours
    to a matrix of k_max nearest neighbours which can be fed as input
    to `mst_from_nn`.


    Parameters
    ----------

    nns : list of length n
        Each nns[i] should be a pair of c_contiguous ndarrays.
        An edge {i, nns[i][0][j]} has weight nns[i][1][j].
        Each nns[i][0] is of type int32 and nns[i][1] of type float32
        (for compatibility with nmslib).
    k_max : int
        If k_max > 0, O(n*k_max) space will be reserved for auxiliary data.


    Returns
    -------

    tuple
        `nn_dist`, `nn_ind`; unused elements (last items in each row)
        will be filled with INFINITY and -1, respectively.
    """
    cdef ssize_t n = len(nns)
    cdef np.ndarray[int]   nn_i
    cdef np.ndarray[float] nn_d

    cdef np.ndarray[ssize_t,ndim=2] ret_nn_ind  = np.empty((n, k_max), dtype=np.intp)
    cdef np.ndarray[float,ndim=2]  ret_nn_dist = np.empty((n, k_max), dtype=np.float32)

    cdef ssize_t i, j, k, l
    cdef ssize_t i1, i2
    cdef float d

    for i in range(n):
        nn_i = nns[i][0]
        nn_d = nns[i][1].astype(np.float32, copy=False)
        k = nn_i.shape[0]
        if nn_d.shape[0] != k:
            raise ValueError("nns has arrays of different lengths as elements")

        l = 0
        for j in range(k):
            i2 = nn_i[j]
            d = nn_d[j]
            if i2 >= 0 and i != i2:
                if l >= k_max: raise ValueError("`k_max` is too small")
                ret_nn_ind[i, l]  = i2
                ret_nn_dist[i, l] = d
                if l > 0 and ret_nn_dist[i, l] < ret_nn_dist[i, l-1]:
                    raise ValueError("nearest neighbours not sorted")
                l += 1

        while l < k_max:
            ret_nn_ind[i, l]  = -1
            ret_nn_dist[i, l] = INFINITY
            l += 1

    return ret_nn_dist, ret_nn_ind



# cpdef tuple mst_from_nn_list(list nns,
#         ssize_t k_max=0,
#         bint stop_disconnected=True,
#         bint verbose=False):
#     """
#     Computes a minimum spanning tree of a (<=k)-nearest neighbour graph
#     using Kruskal's algorithm, and orders its edges w.r.t. increasing weights.
#
#     See `mst_from_nn` for more details.
#
#
#     Parameters
#     ----------
#
#     nns : list of length n
#         Each nns[i] should be a pair of c_contiguous ndarrays.
#         An edge {i, nns[i][0][j]} has weight nns[i][1][j].
#         Each nns[i][0] is of type int32 and nns[i][1] of type float32
#         (for compatibility with nmslib).
#     k_max : int
#         If k_max > 0, O(n*k_max) space will be reserved for auxiliary data.
#     stop_disconnected : bool
#         raise an exception if the input graph is not connected
#     verbose: bool
#         whether to print diagnostic messages
#
#     Returns
#     -------
#
#     pair : tuple
#         See `mst_from_nn`.
#     """
#     cdef ssize_t n = len(nns)
#     cdef np.ndarray[int]   nn_i
#     cdef np.ndarray[float] nn_d
#     cdef ssize_t k
#     cdef ssize_t i, j
#     cdef ssize_t i1, i2
#     cdef float d
#
#     cdef vector[ c_mst.CMstTriple[float] ] nns2
#     if k_max > 0:
#         nns2.reserve(n*k_max)
#
#
#     for i in range(n):
#         nn_i = nns[i][0]
#         nn_d = nns[i][1]
#         k = nn_i.shape[0]
#         if nn_d.shape[0] != k:
#             raise ValueError("nns has arrays of different lengths as elements")
#
#         for j in range(k):
#             i1 = i
#             i2 = nn_i[j]
#             d = nn_d[j]
#             if i2 >= 0 and i1 != i2:
#                 nns2.push_back( c_mst.CMstTriple[float](i1, i2, d) )
#
#     cdef np.ndarray[ssize_t,ndim=2] mst_ind  = np.empty((n-1, 2), dtype=np.intp)
#     cdef np.ndarray[float]          mst_dist = np.empty(n-1, dtype=np.float32)
#
#     cdef ssize_t n_edges = c_mst.Cmst_from_nn_list(nns2.data(), nns2.size(), n,
#             &mst_dist[0], &mst_ind[0,0], verbose)
#
#     if stop_disconnected and n_edges < n-1:
#         raise ValueError("graph is disconnected")
#
#     return mst_dist, mst_ind
#



cpdef tuple mst_from_nn(
    floatT[:,::1] dist,
    ssize_t[:,::1] ind,
    floatT[::1] d_core=None,
    bint stop_disconnected=True,
    bint stop_inexact=False,
    bint verbose=False):
    """Computes a minimum spanning tree(*) of a (<=k)-nearest neighbour graph
    using Kruskal's algorithm, and orders its edges w.r.t. increasing weights.

    Note that in general, the sum of weights in an MST of the (<=k)-nearest
    neighbour graph might be greater than the sum of weights in a minimum
    spanning tree of the complete pairwise distances graph.

    (*) or forest, if the input graph is unconnected. However,
    if stop_disconnected is True, an exception is raised when there is
    no tree spanning a given (<=k)-nn graph.


    Parameters
    ----------

    dist : a c_contiguous ndarray, shape (n,k)
        dist[i,:] is sorted nondecreasingly for all i,
        dist[i,j] gives the weight of the edge {i, ind[i,j]}
    ind : a c_contiguous ndarray, shape (n,k)
        edge definition, interpreted as {i, ind[i,j]}
    d_core : c_contiguous ndarray of length n; optional (default=None)
        core distances for computing the mutual reachability distance
    stop_disconnected : bool
        raise an exception if the input graph is not connected
    verbose: bool
        whether to print diagnostic messages

    Returns:
    -------

    pair : tuple
        A pair (mst_dist, mst_ind) defining the n-1 edges of the MST:
          a) the (n-1)-ary array mst_dist is such that
          mst_dist[i] gives the weight of the i-th edge;
          b) mst_ind is a matrix with n-1 rows and 2 columns,
          where {mst[i,0], mst[i,1]} defines the i-th edge of the tree.

        The results are ordered w.r.t. nondecreasing weights.
          (and then the 1st, and the the 2nd index).
          For each i, it holds mst[i,0]<mst[i,1].

        If stop_disconnected is False, then the weights of the
        last c-1 edges are set to infinity and the corresponding indices
        are set to -1, where c is the number of connected components
        in the resulting minimum spanning forest.
    """
    cdef ssize_t n = dist.shape[0]
    cdef ssize_t k = dist.shape[1]

    if not (ind.shape[0] == n and ind.shape[1] == k):
        raise ValueError("shapes of dist and ind must match")

    cdef np.ndarray[ssize_t,ndim=2] mst_ind  = np.empty((n-1, 2), dtype=np.intp)
    cdef np.ndarray[floatT]         mst_dist = np.empty(n-1,
        dtype=np.float32 if floatT is float else np.float64)

    cdef bool maybe_inexact

    cdef floatT* d_core_ptr = NULL
    if d_core is not None:
        if not (d_core.shape[0] == n):
            raise ValueError("shapes of dist and d_core must match")
        d_core_ptr = &d_core[0]

    _openmp_set_num_threads()
    cdef ssize_t n_edges = c_mst.Cmst_from_nn(
        &dist[0,0], &ind[0,0],
        d_core_ptr,
        n, k,
        &mst_dist[0], &mst_ind[0,0], &maybe_inexact, verbose)

    if stop_disconnected and n_edges < n-1:
        raise ValueError("graph is disconnected")

    if stop_inexact and maybe_inexact:
        raise ValueError("MST maybe inexact")

    return mst_dist, mst_ind




cpdef tuple mst_from_complete(
    floatT[:,::1] X,
    bint verbose=False): # [:,::1]==c_contiguous
    """A Jarník (Prim/Dijkstra)-like algorithm for determining
    a(*) minimum spanning tree (MST) of a complete undirected graph
    with weights given by a symmetric n*n matrix
    or a distance vector of length n*(n-1)/2.

    (*) Note that there might be multiple minimum trees spanning a given graph.


    References:
    ----------

    [1] M. Gagolewski, M. Bartoszuk, A. Cena,
    Genie: A new, fast, and outlier-resistant hierarchical clustering algorithm,
    Information Sciences 363 (2016) 8–23.

    [2] V. Jarník, O jistém problému minimálním,
    Práce Moravské Přírodovědecké Společnosti 6 (1930) 57–63.

    [3] C.F. Olson, Parallel algorithms for hierarchical clustering,
    Parallel Comput. 21 (1995) 1313–1325.

    [4] R. Prim, Shortest connection networks and some generalizations,
    Bell Syst. Tech. J. 36 (1957) 1389–1401.


    Parameters
    ----------

    X : c_contiguous ndarray, shape (n*(n-1)/2, 1) or (n,n)
        distance vector or matrix
    verbose: bool
        whether to print diagnostic messages

    Returns
    -------

    pair : tuple
        A pair (mst_dist, mst_ind) defining the n-1 edges of the MST:
          a) the (n-1)-ary array mst_dist is such that
          mst_dist[i] gives the weight of the i-th edge;
          b) mst_ind is a matrix with n-1 rows and 2 columns,
          where {mst[i,0], mst[i,1]} defines the i-th edge of the tree.

        The results are ordered w.r.t. nondecreasing weights.
        (and then the 1st, and the the 2nd index).
        For each i, it holds mst[i,0]<mst[i,1].
    """
    cdef ssize_t d = X.shape[1]
    cdef ssize_t n = X.shape[0]
    if d == 1:
        n = <ssize_t>libc.math.round((libc.math.sqrt(1.0+8.0*n)+1.0)/2.0)
        assert n*(n-1)//2 == X.shape[0]

    cdef np.ndarray[ssize_t,ndim=2] mst_ind  = np.empty((n-1, 2), dtype=np.intp)
    cdef np.ndarray[floatT]         mst_dist = np.empty(n-1,
        dtype=np.float32 if floatT is float else np.float64)

    cdef c_mst.CDistance[floatT]* D = NULL
    if d == 1:
        D = <c_mst.CDistance[floatT]*>new c_mst.CDistancePrecomputedVector[floatT](&X[0,0], n)
    else:
        assert d == n
        D = <c_mst.CDistance[floatT]*>new c_mst.CDistancePrecomputedMatrix[floatT](&X[0,0], n)

    _openmp_set_num_threads()
    c_mst.Cmst_from_complete(D, n, &mst_dist[0], &mst_ind[0,0], verbose)

    if D:  del D

    return mst_dist, mst_ind




cpdef tuple mst_from_distance(
    floatT[:,::1] X,
    str metric="euclidean",
    floatT[::1] d_core=None,
    bint verbose=False):
    """A Jarník (Prim/Dijkstra)-like algorithm for determining
    a(*) minimum spanning tree (MST) of X with respect to a given metric
    (distance). Distances are computed on the fly.
    Memory use: O(n).


    References
    ----------

    [1] M. Gagolewski, M. Bartoszuk, A. Cena,
    Genie: A new, fast, and outlier-resistant hierarchical clustering algorithm,
    Information Sciences 363 (2016) 8–23.

    [2] V. Jarník, O jistém problému minimálním,
    Práce Moravské Přírodovědecké Společnosti 6 (1930) 57–63.

    [3] C.F. Olson, Parallel algorithms for hierarchical clustering,
    Parallel Comput. 21 (1995) 1313–1325.

    [4] R. Prim, Shortest connection networks and some generalizations,
    Bell Syst. Tech. J. 36 (1957) 1389–1401.


    Parameters
    ----------

    X : c_contiguous ndarray, shape (n,d) or,
            if metric == "precomputed", (n*(n-1)/2,1) or (n,n)
        n data points in a feature space of dimensionality d
        or pairwise distances between n points
    metric : string
        one of ``"euclidean"`` (a.k.a. ``"l2"``),
        ``"manhattan"`` (synonyms: ``"cityblock"``, ``"l1"``),
        ``"cosine"`` (a.k.a. ``"cosinesimil"``), or ``"precomputed"``.
        More metrics/distances might be supported in future versions.
    d_core : c_contiguous ndarray of length n; optional (default=None)
        core distances for computing the mutual reachability distance
    verbose: bool
        whether to print diagnostic messages

    Returns
    -------

    pair : tuple
        A pair (mst_dist, mst_ind) defining the n-1 edges of the MST:
          a) the (n-1)-ary array mst_dist is such that
          mst_dist[i] gives the weight of the i-th edge;
          b) mst_ind is a matrix with n-1 rows and 2 columns,
          where {mst[i,0], mst[i,1]} defines the i-th edge of the tree.

        The results are ordered w.r.t. nondecreasing weights.
        (and then the 1st, and the the 2nd index).
        For each i, it holds mst[i,0]<mst[i,1].
    """
    cdef ssize_t n = X.shape[0]
    cdef ssize_t d = X.shape[1]
    if d == 1 and metric == "precomputed":
        n = <ssize_t>libc.math.round((libc.math.sqrt(1.0+8.0*n)+1.0)/2.0)
        assert n*(n-1)//2 == X.shape[0]
    cdef ssize_t i
    cdef np.ndarray[ssize_t,ndim=2] mst_ind  = np.empty((n-1, 2), dtype=np.intp)
    cdef np.ndarray[floatT]         mst_dist = np.empty(n-1,
        dtype=np.float32 if floatT is float else np.float64)
    cdef c_mst.CDistance[floatT]* D = NULL
    cdef c_mst.CDistance[floatT]* D2 = NULL

    if metric == "euclidean" or metric == "l2":
        D = <c_mst.CDistance[floatT]*>new c_mst.CDistanceEuclidean[floatT](&X[0,0], n, d)
    elif metric == "manhattan" or metric == "cityblock" or metric == "l1":
        D = <c_mst.CDistance[floatT]*>new c_mst.CDistanceManhattan[floatT](&X[0,0], n, d)
    elif metric == "cosine" or metric == "cosinesimil":
        D = <c_mst.CDistance[floatT]*>new c_mst.CDistanceCosine[floatT](&X[0,0], n, d)
    elif metric == "precomputed":
        if d == 1:
            D = <c_mst.CDistance[floatT]*>new c_mst.CDistancePrecomputedVector[floatT](&X[0,0], n)
        else:
            assert d == n
            D = <c_mst.CDistance[floatT]*>new c_mst.CDistancePrecomputedMatrix[floatT](&X[0,0], n)
    else:
        raise NotImplementedError("given `metric` is not supported (yet)")

    if d_core is not None:
        D2 = D # must be deleted separately
        D  = <c_mst.CDistance[floatT]*>new c_mst.CDistanceMutualReachability[floatT](&d_core[0], n, D2)

    _openmp_set_num_threads()
    c_mst.Cmst_from_complete(D, n, &mst_dist[0], &mst_ind[0,0], verbose)

    if D:  del D
    if D2: del D2

    return mst_dist, mst_ind






cpdef tuple knn_from_distance(floatT[:,::1] X, ssize_t k,
       str metric="euclidean", floatT[::1] d_core=None, bint verbose=False):
    """Determines the first k nearest neighbours of each point in X,
    with respect to a given metric (distance).
    Distances are computed on the fly.
    Memory use: O(k*n).

    It is assumed that each query point is not its own neighbour.


    Parameters
    ----------

    X : c_contiguous ndarray, shape (n,d) or,
            if metric == "precomputed", (n*(n-1)/2,1) or (n,n)
        n data points in a feature space of dimensionality d
        or pairwise distances between n points
    k : int < n
        number of nearest neighbours
    metric : string
        one of ``"euclidean"`` (a.k.a. ``"l2"``),
        ``"manhattan"`` (synonyms: ``"cityblock"``, ``"l1"``),
        ``"cosine"`` (a.k.a. ``"cosinesimil"``), or ``"precomputed"``.
        More metrics/distances might be supported in future versions.
    d_core : c_contiguous ndarray of length n; optional (default=None)
        core distances for computing the mutual reachability distance
    verbose: bool
        whether to print diagnostic messages

    Returns
    -------

    pair : tuple
        A pair (dist, ind) representing the k-NN graph, where:
            dist : a c_contiguous ndarray, shape (n,k)
                dist[i,:] is sorted nondecreasingly for all i,
                dist[i,j] gives the weight of the edge {i, ind[i,j]},
                i.e., the distance between the i-th point and its j-th NN.
            ind : a c_contiguous ndarray, shape (n,k)
                edge definition, interpreted as {i, ind[i,j]};
                ind[i,j] is the index of the j-th nearest neighbour of i.
    """
    cdef ssize_t n = X.shape[0]
    cdef ssize_t d = X.shape[1]
    if d == 1 and metric == "precomputed":
        n = <ssize_t>libc.math.round((libc.math.sqrt(1.0+8.0*n)+1.0)/2.0)
        assert n*(n-1)//2 == X.shape[0]

    cdef ssize_t i
    cdef np.ndarray[ssize_t,ndim=2] ind  = np.empty((n, k), dtype=np.intp)
    cdef np.ndarray[floatT,ndim=2]  dist = np.empty((n, k),
        dtype=np.float32 if floatT is float else np.float64)
    cdef c_mst.CDistance[floatT]* D = NULL
    cdef c_mst.CDistance[floatT]* D2 = NULL

    if metric == "euclidean" or metric == "l2":
        D = <c_mst.CDistance[floatT]*>new c_mst.CDistanceEuclidean[floatT](&X[0,0], n, d)
    elif metric == "manhattan" or metric == "cityblock" or metric == "l1":
        D = <c_mst.CDistance[floatT]*>new c_mst.CDistanceManhattan[floatT](&X[0,0], n, d)
    elif metric == "cosine" or metric == "cosinesimil":
        D = <c_mst.CDistance[floatT]*>new c_mst.CDistanceCosine[floatT](&X[0,0], n, d)
    elif metric == "precomputed":
        if d == 1:
            D = <c_mst.CDistance[floatT]*>new c_mst.CDistancePrecomputedVector[floatT](&X[0,0], n)
        else:
            assert d == n
            D = <c_mst.CDistance[floatT]*>new c_mst.CDistancePrecomputedMatrix[floatT](&X[0,0], n)
    else:
        raise NotImplementedError("given `metric` is not supported (yet)")

    if d_core is not None:
        D2 = D # must be deleted separately
        D  = <c_mst.CDistance[floatT]*>new c_mst.CDistanceMutualReachability[floatT](&d_core[0], n, D2)

    _openmp_set_num_threads()
    c_mst.Cknn_from_complete(D, n, k, &dist[0,0], &ind[0,0], verbose)

    if D:  del D
    if D2: del D2

    return dist, ind


################################################################################
# Graph pre-processing routines
################################################################################





cpdef np.ndarray[ssize_t] get_graph_node_degrees(ssize_t[:,::1] ind, ssize_t n):
    """Given an adjacency list representing an undirected simple graph over
    vertex set {0,...,n-1}, return an array deg with deg[i] denoting
    the degree of the i-th vertex. For instance, deg[i]==1 marks a leaf node.


    Parameters
    ----------

    ind : ndarray, shape (m,2)
        A 2-column matrix such that {ind[i,0], ind[i,1]} represents
        one of m undirected edges. Negative indices are ignored.
    n : int
        Number of vertices.


    Returns
    -------

    deg : ndarray, shape(n,)
        An integer array of length n.
    """
    cdef ssize_t num_edges = ind.shape[0]
    assert ind.shape[1] == 2
    cdef np.ndarray[ssize_t] deg = np.empty(n, dtype=np.intp)

    _openmp_set_num_threads()
    c_preprocess.Cget_graph_node_degrees(&ind[0,0], num_edges, n, &deg[0])

    return deg







################################################################################
# Noisy k-partition and other post-processing routines
################################################################################


cpdef np.ndarray[ssize_t] merge_boundary_points(
        ssize_t[:,::1] mst_i,
        ssize_t[::1] c,
        ssize_t[:,::1] nn_i,
        ssize_t M):
    """A noisy k-partition post-processing:
    given a k-partition (with noise points included),
    merges all "boundary" noise points with their nearest
    "core" points.


    Parameters
    ----------

    mst_i : c_contiguous array
        See genieclust.mst.mst_from_distance()
    c : c_contiguous array of shape (n_samples,)
        c[i] gives the cluster id
        (in {-1, 0, 1, ..., k-1} for some k) of the i-th object.
        Class -1 denotes the `noise' cluster.
    nn_i : c_contiguous matrix of shape (n_samples,n_neighbors)
        nn_ind[i,:] gives the indices of the i-th point's
        nearest neighbours; -1 indicates a "missing value"
    M : int
        smoothing factor, M>=2


    Returns
    -------

    c : ndarray, shape (n_samples,)
        A new integer vector c with c[i] denoting the cluster
        id (in {-1, 0, ..., k-1}) of the i-th object.
    """
    cdef np.ndarray[ssize_t] cl2 = np.array(c, dtype=np.intp)

    _openmp_set_num_threads()
    c_postprocess.Cmerge_boundary_points(
        &mst_i[0,0], mst_i.shape[0],
        &nn_i[0,0], nn_i.shape[1], M,
        &cl2[0], cl2.shape[0])

    return cl2


cpdef np.ndarray[ssize_t] merge_noise_points(
        ssize_t[:,::1] mst_i,
        ssize_t[::1] c):
    """A noisy k-partition post-processing:
    given a k-partition (with noise points included),
    merges all noise points with their nearest
    clusters.


    Parameters
    ----------

    mst_i : c_contiguous array
        See genieclust.mst.mst_from_distance()
    c : c_contiguous array of shape (n_samples,)
        c[i] gives the cluster id
        (in {-1, 0, 1, ..., k-1} for some k) of the i-th object.
        Class -1 denotes the `noise' cluster.


    Returns
    -------

    c : ndarray, shape (n_samples,)
        A new integer vector c with c[i] denoting the cluster
        id (in {0, ..., k-1}) of the i-th object.
    """
    cdef np.ndarray[ssize_t] cl2 = np.array(c, dtype=np.intp)

    _openmp_set_num_threads()
    c_postprocess.Cmerge_noise_points(
        &mst_i[0,0], mst_i.shape[0],
        &cl2[0], cl2.shape[0])

    return cl2


cpdef dict get_linkage_matrix(ssize_t[::1] links,
                              floatT[::1] mst_d,
                              ssize_t[:,::1] mst_i):
    """

    Parameters
    ----------

    links : ndarray
        see return value of genieclust.internal.genie_from_mst.
    mst_d, mst_i : ndarray
        Minimal spanning tree defined by a pair (mst_i, mst_d),
        see genieclust.mst.


    Returns
    -------

    Z : dict
        A dictionary with 3 keys: children, distances, counts,
        see the description of Z[:,:2], Z[:,2] and Z[:,3], respectively,
        in scipy.cluster.hierarchy.linkage.
    """
    cdef ssize_t n = mst_i.shape[0]+1
    cdef ssize_t i, i1, i2, par, w, num_unused, j

    if not n-1 == mst_d.shape[0]:
        raise ValueError("ill-defined MST")
    if not n-1 == links.shape[0]:
        raise ValueError("ill-defined MST")

    cdef c_gini_disjoint_sets.CGiniDisjointSets ds = \
        c_gini_disjoint_sets.CGiniDisjointSets(n)

    cdef np.ndarray[ssize_t,ndim=2] children_  = np.empty((n-1, 2), dtype=np.intp)
    cdef np.ndarray[floatT]         distances_ = np.empty(n-1,
        dtype=np.float32 if floatT is float else np.float64)
    cdef np.ndarray[ssize_t]        counts_    = np.empty(n-1, dtype=np.intp)
    cdef np.ndarray[ssize_t]        used       = np.zeros(n-1, dtype=np.intp)
    cdef np.ndarray[ssize_t]        ids        = np.empty(n, dtype=np.intp)

    num_unused = n-1
    for i in range(n-1):
        if links[i] < 0: break # no more mst edges
        if links[i] >= n-1: raise ValueError("ill-defined links")
        used[links[i]] += 1
        num_unused -= 1

    for i in range(n):
        ids[i] = i

    w = -1
    for i in range(n-1):
        if i < num_unused:
            # get the next unused edge (links a leaf node)
            while True:
                w += 1
                assert w < n-1
                if not used[w]: break
        else:
            assert 0 <= i-num_unused < n-1
            w = links[i-num_unused]

        assert 0 <= w < n-1
        i1 = mst_i[w, 0]
        i2 = mst_i[w, 1]
        if not 0 <= i1 < n: raise ValueError("ill-defined MST")
        if not 0 <= i2 < n: raise ValueError("ill-defined MST")

        i1 = ds.find(i1)
        i2 = ds.find(i2)
        children_[i, 0] = ids[i1]
        children_[i, 1] = ids[i2]
        par = ds.merge(i1, i2)
        ids[par] = n+i # see scipy.cluster.hierarchy.linkage
        distances_[i] = mst_d[w] if i >= num_unused else 0.0
        counts_[i] = ds.get_count(par)



    # corrections for departures from ultrametricity:
    # distances_ = genieclust.tools.cummin(distances_[::-1])[::-1]
    for i in range(n-2, 0, -1):
        if distances_[i-1] > distances_[i]:
            distances_[i-1] = distances_[i]

    return dict(
        children=children_,
        distances=distances_,
        counts=counts_
    )


################################################################################
# Augmented DisjointSets
################################################################################



cdef class DisjointSets:
    """
    Disjoint Sets (Union-Find)

    Represents a partition of the set {0,1,...,n-1}

    Path compression for find() is implemented,
    but the union() operation is naive (neither
    it is union by rank nor by size),
    see https://en.wikipedia.org/wiki/Disjoint-set_data_structure.
    This is by design, as some other operations in the current
    package rely on the assumption that the parent id of each
    element is always <= than itself.


    Parameters:
    ----------

    n : ssize_t
        The cardinality of the set whose partitions are generated.
    """
    cdef c_disjoint_sets.CDisjointSets ds

    def __cinit__(self, ssize_t n):
        self.ds = c_disjoint_sets.CDisjointSets(n)

    def __len__(self):
        """
        Returns the number of subsets-1,
        a.k.a. how many calls to union() can we still perform.

        Returns:
        -------

        len : ssize_t
            A value in {0,...,n-1}.
        """
        return self.ds.get_k()-1


    cpdef ssize_t get_n(self):
        """
        Returns the number of elements in the set being partitioned.
        """
        return self.ds.get_n()


    cpdef ssize_t get_k(self):
        """
        Returns the current number of subsets.
        """
        return self.ds.get_k()


    cpdef ssize_t find(self, ssize_t x):
        """
        Finds the subset id for a given x.


        Parameters:
        ----------

        x : ssize_t
            An integer in {0,...,n-1}, representing an element to find.


        Returns:
        -------

        parent_x : ssize_t
            The id of the parent of x.
        """
        return self.ds.find(x)


    cpdef ssize_t union(self, ssize_t x, ssize_t y):
        """
        Merges the sets containing given x and y.

        Let px be the parent id of x, and py be the parent id of y.
        If px < py, then the new parent id of py will be set to py.
        Otherwise, px will have py as its parent.

        If x and y are already members of the same subset,
        an exception is thrown.

        Parameters:
        ----------

        x, y : ssize_t
            Integers in {0,...,n-1}, representing elements
            of two sets to merge.


        Returns:
        -------

        parent : ssize_t
            The id of the parent of x or y, whichever is smaller.
        """

        return self.ds.merge(x, y)


    cpdef np.ndarray[ssize_t] to_list(self):
        """
        Get parent ids of all the elements


        Returns:
        -------

        parents : ndarray, shape (n,)
            A list m such that m[x] denotes the (recursive) parent id of x,
            for x=0,1,...,n.
        """
        cdef ssize_t i
        cdef np.ndarray[ssize_t] m = np.empty(self.ds.get_n(), dtype=np.intp)
        for i in range(self.ds.get_n()):
            m[i] = self.ds.find(i)
        return m


    cpdef np.ndarray[ssize_t] to_list_normalized(self):
        """
        Get the normalised elements' membership information.


        Returns:
        -------

        set_ids : ndarray, shape (n,)
            A list m such that m[x] denotes the normalised parent id of x.
            The resulting values are in {0,1,...,k-1}, where k is the current
            number of subsets in the partition.
        """
        cdef ssize_t i, j
        cdef np.ndarray[ssize_t] m = np.empty(self.ds.get_n(), dtype=np.intp)
        cdef np.ndarray[ssize_t] v = np.zeros(self.ds.get_n(), dtype=np.intp)
        cdef ssize_t c = 1
        for i in range(self.ds.get_n()):
            j = self.ds.find(i)
            if v[j] == 0:
                v[j] = c
                c += 1
            m[i] = v[j]-1
        return m


    def to_lists(self):
        """
        Returns a list of lists representing the current partition.
        This is a slow operation. Do you really need this?


        Returns:
        -------

        partition : list of lists
            A list of length k, where k is the current number
            of sets in a partition. Each list element is a list
            with values in {0,...,n-1}
        """
        cdef ssize_t i
        cdef list tou, out

        tou = [ [] for i in range(self.ds.get_n()) ]
        for i in range(self.ds.get_n()):
            tou[self.ds.find(i)].append(i)

        out = []
        for i in range(self.ds.get_n()):
            if tou[i]: out.append(tou[i])

        return out


    def __repr__(self):
        """
        Calls self.to_lists()
        """
        return "DisjointSets("+repr(self.to_lists())+")"





################################################################################
# Disjoint Sets (Union-Find)
# A Python class to represent partitions of the set {0,1,...,n-1} for any n
################################################################################




cdef class GiniDisjointSets():
    """
    Augmented disjoint sets (Union-Find) over {0,1,...,n-1}.
    The class allows to compute the normalised Gini index of the
    distribution of subset sizes, i.e.,
    $$
        G(x_1,\dots,x_k) = \frac{
        \sum_{i=1}^{n-1} \sum_{j=i+1}^n |x_i-x_j|
        }{
        (n-1) \sum_{i=1}^n x_i
        }.
    $$

    For a use case, see: Gagolewski M., Bartoszuk M., Cena A.,
    Genie: A new, fast, and outlier-resistant hierarchical clustering algorithm,
    Information Sciences 363, 2016, pp. 8-23. doi:10.1016/j.ins.2016.05.003


    Parameters:
    ----------

    n : ssize_t
        The cardinality of the set whose partitions are generated.
    """
    cdef c_gini_disjoint_sets.CGiniDisjointSets ds

    def __cinit__(self, ssize_t n):
        self.ds = c_gini_disjoint_sets.CGiniDisjointSets(n)

    def __len__(self):
        """
        Returns the number of subsets-1,
        a.k.a. how many calls to union() can we still perform.

        Returns:
        -------

        len : ssize_t
            A value in {0,...,n-1}.
        """
        return self.ds.get_k()-1


    cpdef ssize_t get_n(self):
        """
        Returns the number of elements in the set being partitioned.
        """
        return self.ds.get_n()


    cpdef ssize_t get_k(self):
        """
        Returns the current number of subsets.
        """
        return self.ds.get_k()


    cpdef double get_gini(self):
        """
        Returns the Gini index of the distribution of subsets' sizes.

        Run time: O(1), as the Gini index is updated during a call
        to union().
        """
        return self.ds.get_gini()


    cpdef ssize_t get_count(self, ssize_t x):
        """
        Returns the size of the subset containing x.

        Run time: the cost of find(x)
        """
        return self.ds.get_count(x)


    cpdef ssize_t get_smallest_count(self):
        """
        Returns the size of the smallest subset.

        Run time: O(1)
        """
        return self.ds.get_smallest_count()


    cpdef ssize_t find(self, ssize_t x):
        """
        Finds the subset id for a given x.


        Parameters:
        ----------

        x : int
            An integer in {0,...,n-1}, representing an element to find.


        Returns:
        -------

        parent_x : int
            The id of the parent of x.
        """
        return self.ds.find(x)


    cpdef ssize_t union(self, ssize_t x, ssize_t y):
        """
        Merges the sets containing given x and y.

        Let px be the parent id of x, and py be the parent id of y.
        If px < py, then the new parent id of py will be set to py.
        Otherwise, px will have py as its parent.

        If x and y are already members of the same subset,
        an exception is thrown.

        Update time: pessimistically O(sqrt(n)).


        Parameters:
        ----------

        x, y : int
            Integers in {0,...,n-1}, representing elements
            of two sets to merge.


        Returns:
        -------

        parent : int
            The id of the parent of x or y, whichever is smaller.
        """

        return self.ds.merge(x, y)



    cpdef np.ndarray[ssize_t] to_list(self):
        """
        Get parent ids of all the elements


        Returns:
        -------

        parents : ndarray, shape (n,)
            A list m such that m[x] denotes the (recursive) parent id of x,
            for x=0,1,...,n.
        """
        cdef ssize_t i
        cdef np.ndarray[ssize_t] m = np.empty(self.ds.get_n(), dtype=np.intp)
        for i in range(self.ds.get_n()):
            m[i] = self.ds.find(i)
        return m


    cpdef np.ndarray[ssize_t] to_list_normalized(self):
        """
        Get the normalized elements' membership information.


        Returns:
        -------

        set_ids : ndarray, shape (n,)
            A list m such that m[x] denotes the normalised parent id of x.
            The resulting values are in {0,1,...,k-1}, where k is the current
            number of subsets in the partition.
        """
        cdef ssize_t i, j
        cdef np.ndarray[ssize_t] m = np.empty(self.ds.get_n(), dtype=np.intp)
        cdef np.ndarray[ssize_t] v = np.zeros(self.ds.get_n(), dtype=np.intp)
        cdef ssize_t c = 1
        for i in range(self.ds.get_n()):
            j = self.ds.find(i)
            if v[j] == 0:
                v[j] = c
                c += 1
            m[i] = v[j]-1
        return m


    def to_lists(self):
        """
        Returns a list of lists representing the current partition.
        This is a slow operation. Do you really need this?

        Returns:
        -------

        partition : list of lists
            A list of length k, where k is the current number
            of sets in a partition. Each list element is a list
            with values in {0,...,n-1}
        """
        cdef ssize_t i
        cdef list tou, out

        tou = [ [] for i in range(self.ds.get_n()) ]
        for i in range(self.ds.get_n()):
            tou[self.ds.find(i)].append(i)

        out = []
        for i in range(self.ds.get_n()):
            if tou[i]: out.append(tou[i])

        return out


    def get_counts(self):
        """
        Generates an array of subsets' sizes.
        The resulting vector is ordered nondecreasingly.

        Run time: O(k), where k is the current number of subsets.
        """
        cdef ssize_t k = self.ds.get_k()
        cdef np.ndarray[ssize_t] out = np.empty(k, dtype=np.intp)
        self.ds.get_counts(&out[0])
        return out


    def __repr__(self):
        """
        Calls self.to_lists()
        """
        return "GiniDisjointSets("+repr(self.to_lists())+")"




#############################################################################
# The Genie+ Clustering Algorithm (internal)
#############################################################################

cpdef dict genie_from_mst(
        floatT[::1] mst_d,
        ssize_t[:,::1] mst_i,
        ssize_t n_clusters=1,
        double gini_threshold=0.3,
        bint noise_leaves=False,
        bint compute_full_tree=True,
        bint compute_all_cuts=False,
        bint new_merge=False):
    """Compute a k-partition based on a precomputed MST.

    The Genie+ Clustering Algorithm (with extensions)

    Gagolewski M., Bartoszuk M., Cena A.,
    Genie: A new, fast, and outlier-resistant hierarchical clustering algorithm,
    Information Sciences 363, 2016, pp. 8-23. doi:10.1016/j.ins.2016.05.003

    A new hierarchical clustering linkage criterion: the Genie algorithm
    links two clusters in such a way that a chosen economic inequity measure
    (here, the Gini index) of the cluster sizes does not increase drastically
    above a given threshold. The introduced method most often outperforms
    the Ward or average linkage, k-means, spectral clustering,
    DBSCAN, Birch, and others in terms of the clustering
    quality on benchmark data while retaining the single linkage speed.

    This is a new implementation of the O(n sqrt(n))-time version
    of the original algorithm. Additionally, MST leaves can be
    marked as noise points (if `noise_leaves==True`). This is useful,
    if the Genie algorithm is applied on the MST with respect to
    the HDBSCAN-like mutual reachability distance.


    gini_threshold==1.0 and noise_leaves==False, gives the single linkage
    algorithm. Set gini_threshold==1.0 and noise_leaves==True to get
    a HDBSCAN-like behaviour (and make sure
    the MST is computed w.r.t. the mutual reachability distance).


    Parameters
    ----------

    mst_d, mst_i : ndarray
        Minimal spanning tree defined by a pair (mst_i, mst_d),
        see genieclust.mst.
    n_clusters : int
        Number of clusters the dataset is split into.
        If `compute_full_tree` is False, then only partial cluster hierarchy
        is determined.
    gini_threshold : float
        The threshold for the Genie correction
    noise_leaves : bool
        Mark leaves as noise;
        Prevents forming singleton-clusters.
    compute_full_tree : bool
        Compute the whole merge sequence or stop early?
    compute_all_cuts : bool
        Compute the n_clusters and all the more coarse-grained ones?
        Implies `compute_full_tree`.
    new_merge : bool
        False for compatibility with the original Genie algorithm
        (R package `genie`). True (EXPERIMENTAL) merges pairs that
        lower the Gini index below `gini_threshold`
        (if that is possible) -- slower and perhaps not that awesome
        (TODO: testing required).


    Returns
    -------

    res : dict, with the following elements:
        labels : ndarray, shape (n,) or (n_clusters, n) or None
            Is None if n_clusters==0.

            If compute_all_cuts==False, this gives the predicted labels,
            representing an n_clusters-partition of X.
            labels[i] gives the cluster id of the i-th input point.
            If noise_leaves==True, then label -1 denotes a noise point.

            If compute_all_cuts==True, then
            labels[i,:] gives the (i+1)-partition, i=0,...,n_cluster-1.

        links : ndarray, shape (n-1,)
            links[i] gives the MST edge merged at the i-th iteration
            of the algorithm.

        iters : int
            number of merge steps performed

        n_cluster : integer
            actual number of clusters found, 0 if labels is None

    """
    cdef ssize_t n = mst_i.shape[0]+1

    if not 0 <= n_clusters <= n:
        raise ValueError("incorrect n_clusters")
    if not n-1 == mst_d.shape[0]:
        raise ValueError("ill-defined MST")
    if not 0.0 <= gini_threshold <= 1.0:
        raise ValueError("incorrect gini_threshold")


    cdef np.ndarray[ssize_t] tmp_labels_1
    cdef np.ndarray[ssize_t,ndim=2] tmp_labels_2

    cdef np.ndarray[ssize_t] links_  = np.empty(n-1, dtype=np.intp)
    cdef ssize_t n_clusters_ = 0, iters_
    labels_ = None # on request, see below

    _openmp_set_num_threads()

    cdef c_genie.CGenie[floatT] g
    g = c_genie.CGenie[floatT](&mst_d[0], &mst_i[0,0], n, noise_leaves, new_merge)

    if compute_all_cuts:
        compute_full_tree = True

    g.apply_genie(1 if compute_full_tree else n_clusters, gini_threshold)

    iters_ = g.get_links(&links_[0])

    if n_clusters >= 1:
        n_clusters_ = min(g.get_max_n_clusters(), n_clusters)

        if compute_all_cuts:
            tmp_labels_2 = np.empty((n_clusters_, n), dtype=np.intp)
            g.get_labels_matrix(n_clusters_, &tmp_labels_2[0,0])
            labels_ = tmp_labels_2
        else:
            # just one cut:
            tmp_labels_1 = np.empty(n, dtype=np.intp)
            g.get_labels(n_clusters_, &tmp_labels_1[0])
            labels_ = tmp_labels_1

    return dict(labels=labels_,
                n_clusters=n_clusters_,
                links=links_,
                iters=iters_)





#############################################################################
# The Genie+Information Criterion (G+IC) Clustering Algorithm
#############################################################################


cpdef dict gic_from_mst(
        floatT[::1] mst_d,
        ssize_t[:,::1] mst_i,
        double n_features,
        ssize_t n_clusters=1,
        ssize_t add_clusters=0,
        double[::1] gini_thresholds=None,
        bint noise_leaves=False,
        bint compute_full_tree=True,
        bint compute_all_cuts=False):
    """GIc (Genie+Information Criterion) Information-Theoretic
    Hierarchical Clustering Algorithm

    Computes a k-partition based on a pre-computed MST
    maximising (heuristically) the information criterion [2].

    GIc has been proposed by Anna Cena in [1] and was inspired
    by Mueller's (et al.) ITM [2] and Gagolewski's (et al.) Genie [3]

    GIc uses a bottom-up, agglomerative approach (as opposed to the ITM,
    which follows a divisive scheme). It greedily selects for merging
    a pair of clusters that maximises the information criterion [2].
    By default, the initial partition is determined by considering
    the intersection of clusterings found by the Genie methods with
    thresholds 0.1, 0.3, 0.5 and 0.7.


    References
    ==========

    [1] Cena A., Adaptive hierarchical clustering algorithms based on
    data aggregation methods, PhD Thesis, Systems Research Institute,
    Polish Academy of Sciences 2018.

    [2] Mueller A., Nowozin S., Lampert C.H., Information Theoretic
    Clustering using Minimum Spanning Trees, DAGM-OAGM 2012.

    [3] Gagolewski M., Bartoszuk M., Cena A.,
    Genie: A new, fast, and outlier-resistant hierarchical clustering algorithm,
    Information Sciences 363, 2016, pp. 8-23. doi:10.1016/j.ins.2016.05.003



    Parameters
    ----------

    mst_d, mst_i : ndarray
        Minimal spanning tree defined by a pair (mst_i, mst_d),
        see genieclust.mst.
    n_features : double
        number of features in the data set
        [can be fractional if you know what you're doing]
    n_clusters : int
        Number of clusters the dataset is split into.
        If `compute_full_tree` is False, then only partial cluster hierarchy
        is determined.
    add_clusters: int, default=0
        Number of additional clusters to work with internally.
    gini_thresholds : ndarray or None for the default
        Gini index thresholds to use when computing the initial
        partition. Multiple runs of the Genie algorithm with different
        thresholds are explored and the intersection of the resulting
        clusterings is taken as the entry point.
        If gini_thresholds is an empty array, `add_clusters`
        is ignored and the procedure starts from a weak clustering
        (singletons), which we call Agglomerative-IC (ICA).
        If gini_thresholds is of length 1 and add_clusters==0,
        then the procedure is equivalent to the classical Genie algorithm.
    noise_leaves : bool
        Mark leaves as noise;
        Prevents forming singleton-clusters.
    compute_full_tree : bool
        Compute the whole merge sequence or stop early?
        Implies compute_full_tree.
    compute_all_cuts : bool
        Compute the n_clusters and all the more coarse-grained ones?


    Returns
    -------

    res : dict, with the following elements:
        labels : ndarray, shape (n,) or None
            Predicted labels, representing an n_clusters-partition of X.
            labels[i] gives the cluster id of the i-th input point.
            If noise_leaves==True, then label -1 denotes a noise point.
            Is None if n_clusters==0.

        links : ndarray, shape (n-1,)
            links[i] gives the MST edge merged at the i-th iteration
            of the algorithm.

        iters : int
            number of merge steps performed

        n_cluster : integer
            actual number of clusters found, 0 if labels is None
    """
    cdef ssize_t n = mst_i.shape[0]+1

    if not 0 <= n_clusters <= n:
        raise ValueError("incorrect n_clusters")
    if not n-1 == mst_d.shape[0]:
        raise ValueError("ill-defined MST")

    if gini_thresholds is None:
        gini_thresholds = np.r_[0.1, 0.3, 0.5, 0.7]


    cdef np.ndarray[ssize_t] tmp_labels_1
    cdef np.ndarray[ssize_t,ndim=2] tmp_labels_2

    cdef np.ndarray[ssize_t] links_  = np.empty(n-1, dtype=np.intp)
    cdef ssize_t n_clusters_ = 0, iters_
    labels_ = None # on request, see below

    _openmp_set_num_threads()

    cdef c_genie.CGIc[floatT] g
    g = c_genie.CGIc[floatT](&mst_d[0], &mst_i[0,0], n, noise_leaves)

    if compute_all_cuts:
        compute_full_tree = True

    g.apply_gic(1 if compute_full_tree else n_clusters,
                n_clusters-1+add_clusters if compute_full_tree else add_clusters,
                n_features,
            &gini_thresholds[0], gini_thresholds.shape[0])

    iters_ = g.get_links(&links_[0])

    if n_clusters >= 1:
        n_clusters_ = min(g.get_max_n_clusters(), n_clusters)

        if compute_all_cuts:
            tmp_labels_2 = np.empty((n_clusters_, n), dtype=np.intp)
            g.get_labels_matrix(n_clusters_, &tmp_labels_2[0,0])
            labels_ = tmp_labels_2
        else:
            # just one cut:
            tmp_labels_1 = np.empty(n, dtype=np.intp)
            g.get_labels(n_clusters_, &tmp_labels_1[0])
            labels_ = tmp_labels_1

    return dict(labels=labels_,
                n_clusters=n_clusters_,
                links=links_,
                iters=iters_)



# distutils: language=c++
# cython: boundscheck=False
# cython: cdivision=True
# cython: nonecheck=False
# cython: wraparound=False
# cython: language_level=3



"""
Provides access to the Ccontingency_table(), Capply_pivoting()
and Ccompare_partitions_*() functions.

Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License
Version 3, 19 November 2007, published by the Free Software Foundation.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License Version 3 for more details.
You should have received a copy of the License along with this program.
If not, see <https://www.gnu.org/licenses/>.
"""


cdef extern from "../src/c_compare_partitions.h":
    cdef struct CComparePartitionsPairsResult:
        double ar
        double r
        double fm
        double afm

    cdef struct CComparePartitionsInfoResult:
        double mi
        double nmi
        double ami


    void Cminmax[T](const T* x, ssize_t n, T* xmin, T* xmax)

    void Ccontingency_table(ssize_t* C, ssize_t xc, ssize_t yc,
        ssize_t xmin, ssize_t ymin,
        ssize_t* x, ssize_t* y, ssize_t n)

    void Capply_pivoting(ssize_t* C, ssize_t xc, ssize_t yc)

    CComparePartitionsPairsResult Ccompare_partitions_pairs(const ssize_t* C,
        ssize_t xc, ssize_t yc)
    CComparePartitionsInfoResult  Ccompare_partitions_info(const ssize_t* C,
        ssize_t xc, ssize_t yc)
    double Ccompare_partitions_nacc(const ssize_t* C, ssize_t xc, ssize_t yc)
    double Ccompare_partitions_psi(const ssize_t* C, ssize_t xc, ssize_t yc)



"""
The Genie++ Clustering Algorithm
"""


# Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License
# Version 3, 19 November 2007, published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License Version 3 for more details.
# You should have received a copy of the License along with this program.
# If not, see <https://www.gnu.org/licenses/>.


import os
import sys
import math
import numpy as np
import scipy.sparse
from sklearn.base import BaseEstimator, ClusterMixin
from . import internal
import warnings


try:
    import nmslib
except ImportError:
    nmslib = None


try:
    import mlpack
except ImportError:
    mlpack = None


###############################################################################
###############################################################################
###############################################################################


class GenieBase(BaseEstimator, ClusterMixin):
    """
    Base class for Genie and GIc

    For detailed description of the parameters and attributes,
    see `genieclust.Genie`.
    """

    def __init__(
            self,
            *,
            n_clusters,
            M,
            affinity,
            exact,
            compute_full_tree,
            compute_all_cuts,
            postprocess,
            cast_float32,
            mlpack_enabled,
            mlpack_leaf_size,
            nmslib_n_neighbors,
            nmslib_params_init,
            nmslib_params_index,
            nmslib_params_query,
            verbose):
        # # # # # # # # # # # #
        super().__init__()
        self.n_clusters          = n_clusters
        self.n_features          = None  # can be overwritten by GIc
        self.M                   = M
        self.affinity            = affinity
        self.exact               = exact
        self.compute_full_tree   = compute_full_tree
        self.compute_all_cuts    = compute_all_cuts
        self.postprocess         = postprocess
        self.cast_float32        = cast_float32
        self.mlpack_enabled      = mlpack_enabled
        self.mlpack_leaf_size    = mlpack_leaf_size
        self.nmslib_n_neighbors  = nmslib_n_neighbors
        self.nmslib_params_init  = nmslib_params_init
        self.nmslib_params_index = nmslib_params_index
        self.nmslib_params_query = nmslib_params_query
        self.verbose             = verbose

        self.n_samples_           = None
        self.n_features_          = None
        self.n_clusters_          = 0  # should not be confused with self.n_clusters
        self.labels_              = None
        # self.is_noise_            = None
        self.children_            = None
        self.distances_           = None
        self.counts_              = None

        self._mst_dist_           = None
        self._mst_ind_            = None
        self._nn_dist_            = None
        self._nn_ind_             = None
        self._d_core_             = None
        self._links_              = None
        self._iters_              = None

        self._last_state_         = None



    def _postprocess_outputs(self, res, cur_state):
        """
        (internal) Updates `self.labels_`
        """
        if cur_state["verbose"]:
            print("[genieclust] Postprocessing outputs.", file=sys.stderr)

        self.labels_     = res["labels"]
        self._links_     = res["links"]
        self._iters_     = res["iters"]

        if res["n_clusters"] != cur_state["n_clusters"]:
            warnings.warn("The number of clusters detected (%d) is "
                          "different than the requested one (%d)." % (
                            res["n_clusters"],
                            cur_state["n_clusters"]))
        self.n_clusters_ = res["n_clusters"]

        if self.labels_ is not None:
            reshaped = False
            if self.labels_.ndim == 1:
                reshaped = True
                # promote it to a matrix with 1 row
                self.labels_.shape = (1, self.labels_.shape[0])
                start_partition = 0
            else:
                # duplicate the 1st row (create the "0"-partition that will
                # not be postprocessed):
                self.labels_ = np.vstack((self.labels_[0, :], self.labels_))
                start_partition = 1  # do not postprocess the "0"-partition

            #self.is_noise_    = (self.labels_[0, :] < 0)

            # postprocess labels, if requested to do so
            if cur_state["M"] < 2 or cur_state["postprocess"] == "none":
                pass
            elif cur_state["postprocess"] == "boundary":
                assert self._nn_ind_ is not None
                assert self._nn_ind_.shape[1] >= cur_state["M"] - 1
                for i in range(start_partition, self.labels_.shape[0]):
                    self.labels_[i, :] = internal.merge_boundary_points(
                        self._mst_ind_, self.labels_[i, :],
                        self._nn_ind_, cur_state["M"])
            elif cur_state["postprocess"] == "all":
                for i in range(start_partition, self.labels_.shape[0]):
                    self.labels_[i, :] = internal.merge_noise_points(
                        self._mst_ind_, self.labels_[i, :])

        if reshaped:
            self.labels_.shape = (self.labels_.shape[1],)

        if cur_state["compute_full_tree"]:
            assert cur_state["exact"] and cur_state["M"] == 1
            Z = internal.get_linkage_matrix(
                self._links_,
                self._mst_dist_,
                self._mst_ind_)
            self.children_    = Z["children"]
            self.distances_   = Z["distances"]
            self.counts_      = Z["counts"]

        return cur_state


    def _check_params(self, cur_state=None):
        if cur_state is None:
            cur_state = dict()

        cur_state["M"] = int(self.M)
        if cur_state["M"] < 1:
            raise ValueError("`M` must be > 0.")

        cur_state["exact"]             = bool(self.exact)

        cur_state["compute_full_tree"] = bool(self.compute_full_tree)
        if cur_state["compute_full_tree"] and \
                not (cur_state["M"] == 1 and cur_state["exact"]):
            cur_state["compute_full_tree"] = False
            warnings.warn("`compute_full_tree` is only available when `M` = 1 "
                          "and `exact` is True")

        cur_state["compute_all_cuts"]  = bool(self.compute_all_cuts)
        cur_state["cast_float32"]      = bool(self.cast_float32)
        cur_state["verbose"]           = bool(self.verbose)

        cur_state["n_clusters"] = int(self.n_clusters)
        if cur_state["n_clusters"] < 0:
            raise ValueError("n_clusters must be >= 0")

        _postprocess_options = ("boundary", "none", "all")
        cur_state["postprocess"] = str(self.postprocess).lower()
        if cur_state["postprocess"] not in _postprocess_options:
            raise ValueError("`postprocess` should be one of %r" % _postprocess_options)

        cur_state["affinity"] = str(self.affinity).lower()
        if cur_state["affinity"] in ["euclidean", "lp:p=2"]:
            cur_state["affinity"] = "l2"
        elif cur_state["affinity"] in ["euclidean_sparse"]:
            cur_state["affinity"] = "l2_sparse"
        elif cur_state["affinity"] in ["manhattan", "cityblock", "lp:p=1"]:
            cur_state["affinity"] = "l1"
        elif cur_state["affinity"] in ["manhattan_sparse", "cityblock_sparse"]:
            cur_state["affinity"] = "l1_sparse"
        elif cur_state["affinity"] in ["chebyshev", "maximum", "lp:p=inf"]:
            cur_state["affinity"] = "linf"
        elif cur_state["affinity"] in ["chebyshev_sparse", "maximum_sparse"]:
            cur_state["affinity"] = "linf_sparse"
        elif cur_state["affinity"] in ["cosine"]:
            cur_state["affinity"] = "cosinesimil"
        elif cur_state["affinity"] in ["cosine_sparse"]:
            cur_state["affinity"] = "cosinesimil_sparse"
        elif cur_state["affinity"] in ["cosine_sparse_fast"]:
            cur_state["affinity"] = "cosinesimil_sparse_fast"

        _affinity_exact_options = (
            "l2", "euclidean", "l1", "manhattan",
            "cityblock", "cosine", "cosinesimil", "precomputed")
        if cur_state["exact"] and cur_state["affinity"] not in _affinity_exact_options:
            raise ValueError("`affinity` should be one of %r" % _affinity_exact_options)

        if type(self.mlpack_enabled) is str:
            cur_state["mlpack_enabled"] = str(self.mlpack_enabled).lower()
            if cur_state["mlpack_enabled"] != "auto":
                raise ValueError("`mlpack_enabled` must be one of: 'auto', True, False.")
        else:
            cur_state["mlpack_enabled"] = bool(self.mlpack_enabled)

        cur_state["mlpack_leaf_size"] = int(self.mlpack_leaf_size)  # mlpack will check this


        cur_state["nmslib_n_neighbors"] = int(self.nmslib_n_neighbors)

        if type(self.nmslib_params_init) is not dict:
            raise ValueError("`nmslib_params_init` must be a `dict`.")
        cur_state["nmslib_params_init"] = self.nmslib_params_init.copy()

        if type(self.nmslib_params_index) is not dict:
            raise ValueError("`nmslib_params_index` must be a `dict`.")
        cur_state["nmslib_params_index"] = self.nmslib_params_index.copy()

        if type(self.nmslib_params_query) is not dict:
            raise ValueError("`nmslib_params_query` must be a `dict`.")
        cur_state["nmslib_params_query"] = self.nmslib_params_query.copy()

        # this is more like an inherent dimensionality for GIc
        cur_state["n_features"] = self.n_features   # users can set this manually
        if cur_state["n_features"] is not None:      # only GIc needs this
            cur_state["n_features"] = max(1.0, float(cur_state["n_features"]))
        else:
            cur_state["n_features"] = -1

        return cur_state


    def _get_mst_exact(self, X, cur_state):
        if cur_state["affinity"] == "precomputed":
            X = X.reshape(X.shape[0], -1)
            if X.shape[1] not in [1, X.shape[0]]:
                raise ValueError(
                    "`X` must be distance vector "
                    "or a square-form distance matrix, "
                    "see `scipy.spatial.distance.pdist` or "
                    "`scipy.spatial.distance.squareform`.")
            if X.shape[1] == 1:
                # from a very advanced and sophisticated quadratic equation:
                n_samples = int(round((math.sqrt(1.0+8.0*X.shape[0])+1.0)/2.0))
                assert n_samples*(n_samples-1)//2 == X.shape[0]
            else:
                n_samples  = X.shape[0]
        else:
            if cur_state["cast_float32"]:
                if scipy.sparse.isspmatrix(X):
                    raise ValueError("Sparse matrices are (currently) only "
                                     "supported when `exact` is False")
                X = np.array(X, dtype=np.float32, order="C", copy=False, ndmin=2)

            n_samples  = X.shape[0]
            if cur_state["n_features"] < 0:
                cur_state["n_features"] = X.shape[1]

        if cur_state["mlpack_enabled"] == "auto":
            cur_state["mlpack_enabled"] = mlpack is not None and \
                    cur_state["affinity"] == "l2" and \
                    X.shape[1] <= 6 and \
                    cur_state["M"] == 1

        if cur_state["mlpack_enabled"]:
            if mlpack is None:
                raise ValueError("Package `mlpack` is not available.")
            elif cur_state["affinity"] != "l2":
                raise ValueError("`mlpack` can only be used with `affinity` = 'l2'.")
            elif cur_state["M"] != 1:
                raise ValueError("`mlpack` can only be used with `M` = 1.")

        mst_dist = None
        mst_ind  = None
        nn_dist  = None
        nn_ind   = None
        d_core   = None

        if self._last_state_ is not None and \
                cur_state["X"]            == self._last_state_["X"] and \
                cur_state["affinity"]     == self._last_state_["affinity"] and \
                cur_state["exact"]        == self._last_state_["exact"] and \
                cur_state["cast_float32"] == self._last_state_["cast_float32"]:

            if cur_state["M"] == self._last_state_["M"]:
                mst_dist = self._mst_dist_
                mst_ind  = self._mst_ind_
                nn_dist  = self._nn_dist_
                nn_ind   = self._nn_ind_
            elif cur_state["M"] < self._last_state_["M"]:
                nn_dist  = self._nn_dist_
                nn_ind   = self._nn_ind_

        if cur_state["mlpack_enabled"]:
            assert cur_state["M"] == 1
            assert cur_state["affinity"] == "l2"

            if mst_dist is None or mst_ind is None:
                _res = mlpack.emst(
                    input=X,
                    leaf_size=cur_state["mlpack_leaf_size"],
                    naive=False,
                    copy_all_inputs=False,
                    verbose=cur_state["verbose"])["output"]
                mst_dist = _res[:,  2].astype(X.dtype, order="C")
                mst_ind  = _res[:, :2].astype(np.intp, order="C")
        else:
            if cur_state["M"] >= 2:  # else d_core   = None
                # Genie+HDBSCAN --- determine d_core
                # TODO: mlpack for k-nns?
                if nn_dist is None or nn_ind is None:
                    nn_dist, nn_ind = internal.knn_from_distance(
                        X,  # if not c_contiguous, raises an error
                        k=cur_state["M"]-1,
                        metric=cur_state["affinity"],  # supports "precomputed"
                        verbose=cur_state["verbose"])

                d_core = internal.get_d_core(nn_dist, nn_ind, cur_state["M"])

            # Use Prim's algorithm to determine the MST
            # w.r.t. the distances computed on the fly
            if mst_dist is None or mst_ind is None:
                mst_dist, mst_ind = internal.mst_from_distance(
                    X,  # if not c_contiguous, raises an error
                    metric=cur_state["affinity"],
                    d_core=d_core,
                    verbose=cur_state["verbose"])

        self.n_samples_   = n_samples
        self._mst_dist_   = mst_dist
        self._mst_ind_    = mst_ind
        self._nn_dist_    = nn_dist
        self._nn_ind_     = nn_ind
        self._d_core_     = d_core

        return cur_state


    def _get_mst_approx(self, X, cur_state):
        if nmslib is None:
            raise ValueError("Package `nmslib` is not available.")

        if cur_state["affinity"] == "precomputed":
            raise ValueError(
                "`affinity` of 'precomputed' can only be used "
                "with `exact` = True.")

        if cur_state["cast_float32"]:
            if cur_state["affinity"] in ["leven", "normleven", "jaccard_sparse",
                                         "bit_jaccard", "bit_hamming"]:
                raise ValueError("`cast_float32` cannot be used with this `affinity`.")

            if scipy.sparse.isspmatrix(X):
                X = scipy.sparse.csr_matrix(X, dtype=np.float32, copy=False)
            else:
                X = np.array(X, dtype=np.float32, order="C", copy=False, ndmin=2)

        n_samples  = np.shape(X)[0]

        if cur_state["n_features"] < 0 and np.ndim(X) >= 2:
            cur_state["n_features"] = np.shape(X)[1]

        if "indexThreadQty" in cur_state["nmslib_params_index"]:
            warnings.warn("Set `indexThreadQty` via the OMP_NUM_THREADS "
                          "environment variable.")

        if os.getenv("OMP_NUM_THREADS"):
            n_threads = max(1, int(os.getenv("OMP_NUM_THREADS")))
            cur_state["nmslib_params_index"]["indexThreadQty"] = n_threads
        else:
            n_threads = 0

        if "space" in cur_state["nmslib_params_init"]:
            warnings.warn("Set `space` via the `affinity` parameter.")
        cur_state["nmslib_params_init"]["space"] = cur_state["affinity"]

        if "data_type" not in cur_state["nmslib_params_init"]:
            # nmslib.DataType.DENSE_VECTOR|OBJECT_AS_STRING|SPARSE_VECTOR
            if scipy.sparse.isspmatrix(X):
                data_type = nmslib.DataType.SPARSE_VECTOR
            elif np.ndim(X) == 2:
                data_type = nmslib.DataType.DENSE_VECTOR
            else:
                data_type = nmslib.DataType.OBJECT_AS_STRING
            cur_state["nmslib_params_init"]["data_type"] = data_type

        if "dtype" not in cur_state["nmslib_params_init"]:
            # nmslib.DistType.FLOAT|INT use FLOAT except for `leven`
            cur_state["nmslib_params_init"]["dtype"] = \
                nmslib.DistType.FLOAT if cur_state["affinity"] != "leven" else \
                nmslib.DistType.INT

        cur_state["nmslib_n_neighbors"] = min(
            n_samples-1,
            max(1, cur_state["nmslib_n_neighbors"]))

        if cur_state["nmslib_n_neighbors"] < cur_state["M"]-1:
            raise ValueError("Increase `nmslib_n_neighbors` or decrease `M`.")

        mst_dist = None
        mst_ind  = None
        nn_dist  = None
        nn_ind   = None
        d_core   = None


        if self._last_state_ is not None and \
                cur_state["X"]                   == self._last_state_["X"] and \
                cur_state["affinity"]            == self._last_state_["affinity"] and \
                cur_state["exact"]               == self._last_state_["exact"] and \
                cur_state["nmslib_n_neighbors"]  == self._last_state_["nmslib_n_neighbors"] and \
                cur_state["nmslib_params_init"]  == self._last_state_["nmslib_params_init"] and \
                cur_state["nmslib_params_index"] == self._last_state_["nmslib_params_index"] and \
                cur_state["nmslib_params_query"] == self._last_state_["nmslib_params_query"]:

            if cur_state["M"] == self._last_state_["M"]:
                mst_dist = self._mst_dist_
                mst_ind  = self._mst_ind_
                nn_dist  = self._nn_dist_
                nn_ind   = self._nn_ind_
            elif cur_state["M"] < self._last_state_["M"]:
                nn_dist  = self._nn_dist_
                nn_ind   = self._nn_ind_

        if nn_dist is None or nn_ind is None:
            index = nmslib.init(**cur_state["nmslib_params_init"])
            index.addDataPointBatch(X)
            index.createIndex(
                cur_state["nmslib_params_index"],
                print_progress=cur_state["verbose"])
            index.setQueryTimeParams(cur_state["nmslib_params_query"])
            nns = index.knnQueryBatch(
                X,
                k=cur_state["nmslib_n_neighbors"] + 1,  # will return self as well
                num_threads=n_threads)
            index = None  # no longer needed
            nn_dist, nn_ind = internal.nn_list_to_matrix(nns, cur_state["nmslib_n_neighbors"] + 1)
            nns = None  # no longer needed

        if cur_state["M"] > 1:
            d_core = internal.get_d_core(nn_dist, nn_ind, cur_state["M"])


        if mst_dist is None or mst_ind is None:
            mst_dist, mst_ind = internal.mst_from_nn(
                nn_dist,
                nn_ind,
                d_core,
                stop_disconnected=False,
                verbose=cur_state["verbose"])
            # We can have a forest here...

        self.n_samples_   = n_samples
        self._mst_dist_   = mst_dist
        self._mst_ind_    = mst_ind
        self._nn_dist_    = nn_dist
        self._nn_ind_     = nn_ind
        self._d_core_     = d_core

        return cur_state


    def _get_mst(self, X, cur_state):
        cur_state["X"] = id(X)

        if cur_state["verbose"]:
            print("[genieclust] Preprocessing data.", file=sys.stderr)

        if cur_state["exact"]:
            cur_state = self._get_mst_exact(X, cur_state)
        else:
            cur_state = self._get_mst_approx(X, cur_state)

        # this might be an "intrinsic" dimensionality:
        self.n_features_  = cur_state["n_features"]
        self._last_state_ = cur_state  # will be modified in-place further on

        return cur_state


    def fit_predict(self, X, y=None):
        """
        Perform cluster analysis of a dataset and return the predicted labels.


        Parameters
        ----------

        X : object
            See `genieclust.Genie.fit`.

        y : None
            See `genieclust.Genie.fit`.


        Returns
        -------

        labels_ : ndarray
            `self.labels_` attribute.


        See also
        --------

        genieclust.Genie.fit

        """
        self.fit(X)
        return self.labels_





###############################################################################
###############################################################################
###############################################################################



class Genie(GenieBase):
    """
    Genie++ hierarchical clustering algorithm


    Parameters
    ----------

    n_clusters : int
        Number of clusters to detect.

        If `M` > 1 and `postprocess` is not ``"all"``, `n_clusters` = 1
        can act as a noise point/outlier detector.
        The approximate method (see parameter `exact`) can sometimes
        fail to detect the coarsest-grained partitions; in such a case,
        more clusters might be returned (with a warning).

    gini_threshold : float
        Threshold for the Genie correction in [0,1].

        The Gini index is used to quantify the inequality of the cluster
        size distribution. Low thresholds highly penalise the formation
        of small clusters. Threshold of 1.0 disables the correction
        and for `M` = 1 makes the method be equivalent to the single
        linkage algorithm.

        The algorithm tends to be *stable* with respect to small changes
        to the threshold — they do not tend to affect the output clustering.
        Usually, thresholds of 0.1, 0.3, 0.5, and 0.7 are worth giving a try.

    M : int
        Smoothing factor for the mutual reachability distance [6]_.

        `M` = 1 gives the original Genie algorithm  [1]_ (with no noise point
        detection) with respect to the chosen affinity as-is. Note that
        for `M > 1` we need additionally :math:`O(M n)` working memory
        for storing of points' nearest neighbours.

    affinity : str
        Metric used to compute the linkage.

        *   For `exact` = ``True``:

            One of:
            ``"l2"`` (synonym: ``"euclidean"``),
            ``"l1"`` (synonym: ``"manhattan"``, ``"cityblock"``),
            ``"cosinesimil"`` (synonym: ``"cosine"``), or
            ``"precomputed"``.

            In the latter case, the `X` argument to the `fit` method
            must be a distance vector or a square-form distance matrix, see `scipy.spatial.distance.pdist`.

        *   For `exact` = ``False``:

            Any dissimilarity supported by `nmslib`, see [5]_ and
            https://github.com/nmslib/nmslib/blob/master/manual/spaces.md,
            for instance:
            ``"l2"``,
            ``"l2_sparse"``,
            ``"l1"``,
            ``"l1_sparse"``,
            ``"linf"``,
            ``"linf_sparse"``,
            ``"cosinesimil"``,
            ``"cosinesimil_sparse"``,
            ``"negdotprod"``,
            ``"negdotprod_sparse"``,
            ``"angulardist"``,
            ``"angulardist_sparse"``,
            ``"leven"``,
            ``"normleven"``,
            ``"jaccard_sparse"``,
            ``"bit_jaccard"``,
            ``"bit_hamming"``.

    exact : bool
        Whether to compute the minimum spanning tree exactly or rather
        estimate it based on an approximate near-neighbour graph.

        The exact method has time complexity of :math:`O(d n^2)` [2]_
        (however, see `mlpack_enabled`) but only needs :math:`O(n)` memory.

        If `exact` is ``False``, the minimum spanning tree is approximated
        based on an approximate :math:`k`\\ -nearest neighbours graph found by
        `nmslib` [5]_. This is typically very fast but requires
        :math:`O(k n)` memory.

    compute_full_tree : bool
        Whether to determine the whole cluster hierarchy and the
        linkage matrix.

        Only available if `M` = 1 and `exact` is ``True``.
        Enables plotting of dendrograms or cutting
        of the hierarchy at an arbitrary level, see the
        `children_`, `distances_`, `counts_` attributes.

    compute_all_cuts : bool
        Whether to compute the requested `n_clusters`-partition and all
        the coarser-grained ones.

        If ``True``, then the `labels_` attribute will be a matrix, see below.

    postprocess : {"boundary", "none", "all"}
        Controls the treatment of noise points after the clusters are
        identified.

        In effect only if `M` > 1. Each leaf in the minimum spanning tree
        is treated as a *noise* point. We call it a *boundary point*
        if it is amongst its adjacent vertex's `M` - 1 nearest neighbours.
        By default, only boundary points are merged with their nearest
        *core* points.

        To force a classical `n_clusters`-partition
        of a data set (with no notion of noise), choose ``"all"``.
        Furthermore, ``"none"`` leaves all leaves, i.e., noise points
        (including the boundary ones) as-is.


    cast_float32 : bool
        Whether casting of data type to ``float32`` is to be performed.

        If `exact` is ``True``, it decreases the run-time ca. 2 times
        at a cost of greater memory use. Otherwise, note that `nmslib`
        *requires* ``float32`` data anyway when using dense or sparse
        numeric matrix inputs.

        By setting `cast_float32` to ``False`` a user assures themself
        that the inputs are of acceptable form.

    mlpack_enabled : "auto" or bool
        Whether `mlpack.emst` should be used for computing the Euclidean
        minimum spanning tree instead of the Jarník-Prim algorithm
        when `exact` is ``True``.

        Often fast for very low-dimensional spaces. As the name suggests,
        only `affinity` of ``'l2'`` is supported (and `M` = 1).
        By default, we rely on `mlpack` if it is installed and
        `n_features` <= 6.

    mlpack_leaf_size : int
        Leaf size in the kd-tree when `mlpack.emst` is used.

        According to the `mlpack` manual, leaves of size 1 give the best
        performance at the cost of greater memory use.

    nmslib_n_neighbors : int
        The number of approximate nearest neighbours used to estimate
        the minimum spanning tree when when `exact` is ``False``.

        If the number of nearest neighbours is too small, the nearest
        neighbour graph might be disconnected and the number of obtained
        clusters might be greater than the requested one.

        `nmslib_n_neighbors` must not be less than `M` - 1.

    nmslib_params_init : dict
        A dictionary of parameters to be passed to `nmslib.init`
        when `exact` is ``False``.

        See https://github.com/nmslib/nmslib/blob/master/manual/methods.md,
        https://github.com/nmslib/nmslib/blob/master/manual/spaces.md
        and https://nmslib.github.io/nmslib/
        for more details. The `space`, `data_type`, and `dtype` parameters
        will be set based on the chosen `affinity` and the input `X`.

    nmslib_params_index : dict
        A dictionary of parameters to be passed to `index.createIndex`,
        where `index` is the object constructed with `nmslib.init`.

        The `indexThreadQty` parameter will be set based on the
        ``OMP_NUM_THREADS`` environment variable.

    nmslib_params_query : dict
        A dictionary of parameters to be passed to `index.setQueryTimeParams`,
        where `index` is the object constructed with `nmslib.init`.

    verbose : bool
        Whether to print diagnostic messages and progress information
        on ``stderr``.


    Attributes
    ----------

    labels_ : ndarray
        Detected cluster labels.

        If `compute_all_cuts` is ``False`` (the default),
        this is an integer vector such that ``labels_[i]`` gives
        the cluster ID (between 0 and `n_clusters_` - 1) of the i-th object.
        If `M` > 1, noise points are labelled -1 (unless taken care of in the
        postprocessing stage).

        Otherwise, i.e., if `compute_all_cuts` is ``True``,
        all partitions of cardinality down to `n_clusters`
        are determined; ``labels_[j,i]`` denotes the cluster ID of the i-th
        point in a j-partition. We assume that both the 0- and 1- partition
        distinguishes only between noise- and non-noise points,
        however, no postprocessing is conducted on the 0-partition
        (there might be points with labels of -1 even if `postprocess`
        is ``"all"``).

        Note that the approximate method (`exact` of ``False``) might fail
        to determine the fine-grained clusters (if the approximate
        neighbour graph is disconnected) - the actual number of clusters
        detected can be larger.

    n_clusters_ : int
        The actual number of clusters detected by the algorithm.

        As we argued above, the approximate method might sometimes yield
        a more fine-grained partition than the requested one (with a warning).
        Moreover, there might be too many noise points in the dataset.

    n_samples_ : int
        The number of points in the fitted dataset.

    n_features_ : int
        The number of features in the fitted dataset.

        If the information is not available, it is be set to -1.

    children_ : None or ndarray
        If `compute_full_tree` is ``True``, this is a matrix whose
        i-th row provides the information on the clusters merged in
        the i-th iteration. See the description of ``Z[:,0]`` and ``Z[:,1]``
        in `scipy.cluster.hierarchy.linkage`. Together with `distances_` and
        `counts_`, this constitutes the linkage matrix that can be used for
        plotting the dendrogram.

    distances_ : None or ndarray
        If `compute_full_tree` is ``True``, this is a vector that gives
        the distance between two clusters merged in each iteration,
        see the description of ``Z[:,2]`` in `scipy.cluster.hierarchy.linkage`.

        As the original Genie algorithm does not guarantee that that distances
        are ordered increasingly (there are other hierarchical
        clustering linkages that violate the ultrametricity property as well),
        these are corrected by applying
        ``distances_ = genieclust.tools.cummin(distances_[::-1])[::-1]``.

    counts_ : None or ndarray
        If `compute_full_tree` is ``True``, this is a vector giving
        the number of elements in a cluster created in each iteration.
        See the description of ``Z[:,3]`` in `scipy.cluster.hierarchy.linkage`.



    Notes
    -----

    **Genie** is a robust and outlier resistant
    hierarchical clustering algorithm [1]_, originally published
    as an R package ``genie``. This new implementation is, amongst others,
    much faster and now features optional smoothing and noise
    point detection (if `M` > 1).

    Genie is based on a minimum spanning tree (MST) of the
    pairwise distance graph of a given point set.
    Just like the single linkage, it consumes the edges
    of the MST in increasing order of weights. However, it prevents
    the formation of clusters of highly imbalanced sizes; once the Gini index
    of the cluster size distribution raises above an assumed threshold,
    a forced merge of a point group of the smallest size is performed.
    Its appealing simplicity goes hand in hand with its usability;
    Genie often outperforms other clustering approaches on benchmark data.

    The Genie algorithm itself has :math:`O(n \\sqrt{n})` time
    and :math:`O(n)` memory complexity provided that a minimum spanning
    tree of the pairwise distance graph is given.
    Generally, our parallelised implementation of a Jarník (Prim/Dijkstra)-like
    method [2]_ will be called to compute an MST, which takes :math:`O(d n^2)` time.
    However, `mlpack` [3]_ provides a very fast
    alternative in the case of Euclidean spaces of (very) low dimensionality
    and `M` = 1, see [4]_ and the `mlpack_enabled` parameter.
    Moreover, in the approximate method (`exact` = ``False``) we apply
    the Kruskal algorithm on the near-neighbour graph determined
    by `nmslib` [5]_. Albeit this only gives *some* sort of a spanning *forest*,
    such a data structure turns out to be very suitable for our clustering task
    (note that the set of connected components will determine the top
    level of the identified cluster hierarchy).

    The Genie correction together with the smoothing factor `M` > 1
    gives a robustified version of the HDBSCAN\\* [6]_ algorithm that,
    contrary to its predecessor, is able to detect a *predefined* number of
    clusters. Hence, it is independent of the *DBSCAN*'s somehow magical
    ``eps`` parameter or the *HDBSCAN*'s ``min_cluster_size`` one.
    If `M` > 1, then the minimum spanning tree is computed with respect to the
    mutual reachability distance (based, e.g., on the Euclidean metric).
    Formally, the distance :math:`m(i,j)` is used instead of the
    chosen "raw" distance, :math:`d(i,j)`. It holds
    :math:`m(i,j)=\\max(d(i,j), c(i), c(j))`,
    where the "core" distance :math:`c(i)` is given by
    :math:`d(i,k)` with :math:`k` being the (`M` - 1)-th nearest neighbour
    of :math:`i`. This makes "noise" and "boundary" points being "pulled away"
    from each other, however, note that `M` = 2 corresponds to the
    original distance. During the clustering procedure, all leaves of the MST
    do not take part in the clustering process. They may be merged
    with the nearest clusters during the postprocessing stage,
    or left marked as "noise" observations.


    :Environment variables:
        OMP_NUM_THREADS
            Controls the number of threads used for computing the minimum
            spanning tree (not supported by `mlpack`).


    References
    ----------

    .. [1]
        Gagolewski M., Bartoszuk M., Cena A.,
        Genie: A new, fast, and outlier-resistant hierarchical
        clustering algorithm, *Information Sciences* 363, 2016, 8-23.
        doi:10.1016/j.ins.2016.05.003.

    .. [2] Olson C.F., Parallel algorithms for hierarchical clustering,
        *Parallel Computing* 21(8), 1995, 1313-1325.
        doi:10.1016/0167-8191(95)00017-I.

    .. [3]
        Curtin R.R., Edel M., Lozhnikov M., Mentekidis Y., Ghaisas S., Zhang S.,
        mlpack 3: A fast, flexible machine learning library,
        *Journal of Open Source Software* 3(26), 726, 2018.
        doi:10.21105/joss.00726.

    .. [4]
        March W.B., Ram P., Gray A.G.,
        Fast Euclidean Minimum Spanning Tree: Algorithm, Analysis, and Applications,
        *Proc. ACM SIGKDD'10*, 2010, 603-611.

    .. [5]
        Naidan B., Boytsov L., Malkov Y.,  Novak D.,
        *Non-metric space library (NMSLIB) manual*, version 2.0, 2019.
        https://github.com/nmslib/nmslib/blob/master/manual/latex/manual.pdf.

    .. [6]
        Campello R., Moulavi D., Zimek A., Sander J.,
        Hierarchical density estimates for data clustering, visualization,
        and outlier detection,
        *ACM Transactions on Knowledge Discovery from Data* 10(1),
        2015, 5:1–5:51. doi:10.1145/2733381.

    """

    def __init__(
            self,
            n_clusters=2,
            *,
            gini_threshold=0.3,
            M=1,
            affinity="l2",
            exact=True,
            compute_full_tree=False,
            compute_all_cuts=False,
            postprocess="boundary",
            cast_float32=True,
            mlpack_enabled="auto",
            mlpack_leaf_size=1,
            nmslib_n_neighbors=64,
            nmslib_params_init=dict(method="hnsw"),
            nmslib_params_index=dict(post=2),
            nmslib_params_query=dict(),
            verbose=False):
        # # # # # # # # # # # #
        super().__init__(
            n_clusters=n_clusters,
            M=M,
            affinity=affinity,
            exact=exact,
            compute_full_tree=compute_full_tree,
            compute_all_cuts=compute_all_cuts,
            postprocess=postprocess,
            cast_float32=cast_float32,
            mlpack_enabled=mlpack_enabled,
            mlpack_leaf_size=mlpack_leaf_size,
            nmslib_n_neighbors=nmslib_n_neighbors,
            nmslib_params_init=nmslib_params_init,
            nmslib_params_index=nmslib_params_index,
            nmslib_params_query=nmslib_params_query,
            verbose=verbose)

        self.gini_threshold = gini_threshold
        self._new_merge = False  # experimental, likely to be removed (#51)

        self._check_params()




    def _check_params(self, cur_state=None):
        cur_state = super()._check_params(cur_state)

        cur_state["gini_threshold"] = float(self.gini_threshold)
        if not (0.0 <= cur_state["gini_threshold"] <= 1.0):
            raise ValueError("`gini_threshold` not in [0,1].")

        cur_state["new_merge"] = bool(self._new_merge)  # experimental (#51)

        return cur_state


    def fit(self, X, y=None):
        """
        Perform cluster analysis of a dataset.


        Parameters
        ----------

        X : object
            Typically a matrix with ``n_samples`` rows and ``n_features``
            columns, see below for more details and options.

        y : None
            Ignored.


        Returns
        -------

        self : genieclust.Genie
            The object that the method was called on.


        Notes
        -----

        Refer to the `labels_` and `n_clusters_` attributes for the result.


        Acceptable `X` types depend whether we use the exact or the approximate
        method.

        *   `X` when `exact` = ``True``.

            For `affinity` of ``"precomputed"``, `X` should either
            be a distance vector of length ``n_samples*(n_samples-1)/2``
            (see `scipy.spatial.distance.pdist`) or a square distance matrix
            of shape ``(n_samples, n_samples)``
            (see `scipy.spatial.distance.squareform`).

            Otherwise, `X` should be real-valued matrix
            (dense ``numpy.ndarray``, or an object coercible to)
            with ``n_samples`` rows and ``n_features`` columns.

            In the latter case, it might be a good idea to standardise
            or at least somehow preprocess the coordinates of the input data
            points by calling, for instance,
            ``X = (X-X.mean(axis=0))/X.std(axis=None, ddof=1)``
            so that the dataset is centred at 0 and has total variance of 1.
            This way the method becomes translation and scale invariant.
            What's more, if data are recorded with small precision (say, up
            to few decimal digits), adding a tiny bit of Gaussian noise will
            assure the solution is unique (note that this generally applies
            to other distance-based clustering algorithms as well).

        *   `X` when `exact` = ``False``.

            The approximate method relies on `nmslib` for locating the
            nearest neighbours. Therefore, it supports all datatypes
            described in https://github.com/nmslib/nmslib/blob/master/manual/spaces.md.
            Depending on the chosen `affinity`, `X` may hence be
            a real-valued ``numpy.ndarray`` matrix with ``n_samples`` rows
            and ``n_features`` columns,
            a ``scipy.sparse.csr_matrix`` object, or
            an array of ASCII strings.



        References
        ----------

        .. [1]
            Naidan B., Boytsov L., Malkov Y.,  Novak D.,
            *Non-metric space library (NMSLIB) manual*, version 2.0, 2019.
            https://github.com/nmslib/nmslib/blob/master/manual/latex/manual.pdf.


        """
        cur_state = self._check_params()  # re-check, they might have changed

        cur_state = self._get_mst(X, cur_state)

        if cur_state["verbose"]:
            print("[genieclust] Determining clusters with Genie++.", file=sys.stderr)

        # apply the Genie++ algorithm (the fast part):
        res = internal.genie_from_mst(
            self._mst_dist_,
            self._mst_ind_,
            n_clusters=cur_state["n_clusters"],
            gini_threshold=cur_state["gini_threshold"],
            noise_leaves=(cur_state["M"] > 1),
            compute_full_tree=cur_state["compute_full_tree"],
            compute_all_cuts=cur_state["compute_all_cuts"],
            new_merge=cur_state["new_merge"])

        cur_state = self._postprocess_outputs(res, cur_state)

        if cur_state["verbose"]:
            print("[genieclust] Done.", file=sys.stderr)

        return self




###############################################################################
###############################################################################
###############################################################################



class GIc(GenieBase):
    """
    (**EXPERIMENTAL**) GIc hierarchical clustering algorithm



    Parameters
    ----------

    n_clusters : int
        See `genieclust.Genie`.

    gini_thresholds : array_like
        A list of Gini index thresholds between 0 and 1.

        The GIc algorithm optimises the information criterion in
        an agglomerative way, starting from the intersection of
        the clusterings returned by
        ``Genie(n_clusters=n_clusters+add_clusters, gini_threshold=gini_thresholds[i])``,
        for all ``i`` from ``0`` to ``len(gini_thresholds)-1``.

    M : int
        See `genieclust.Genie`.

    affinity : str
        See `genieclust.Genie`.

    exact : bool, default=True
        See `genieclust.Genie`.

    compute_full_tree : bool
        See `genieclust.Genie`.

    compute_all_cuts : bool
        See `genieclust.Genie`.

        Note that if `compute_all_cuts` is ``True``, then the i-th cut
        in the hierarchy behaves as if `add_clusters` was equal to
        ``n_clusters-i``. In other words, the returned cuts might be different
        from those obtained by multiple calls to GIc, each time with
        different `n_clusters` and constant `add_clusters` requested.

    postprocess : {"boundary", "none", "all"}
        See `genieclust.Genie`.

    cast_float32 : bool
        See `genieclust.Genie`.

    mlpack_enabled : "auto" or bool
        See `genieclust.Genie`.

    mlpack_leaf_size : int
        See `genieclust.Genie`.

    nmslib_n_neighbors : int
        See `genieclust.Genie`.

    nmslib_params_init : dict
        See `genieclust.Genie`.

    nmslib_params_index : dict
        See `genieclust.Genie`.

    nmslib_params_query : dict
        See `genieclust.Genie`.

    add_clusters : int
        Number of additional clusters to work with internally.

    n_features : float or None
        Dataset's (intrinsic) dimensionality.

        If ``None``, it will be set based on the shape of the input matrix.
        Yet, `affinity` of ``"precomputed"`` needs this to be set manually.

    verbose : bool
        See `genieclust.Genie`.


    Attributes
    ----------

    See class `genieclust.Genie`.



    See also
    --------

    genieclust.Genie



    Notes
    -----

    GIc (Genie+Information Criterion) is an Information-Theoretic
    Hierarchical Clustering Algorithm.
    It was proposed by Anna Cena in [1]_ and had been inspired
    by Mueller's (et al.) ITM [2]_ and Gagolewski's (et al.) Genie [3]_.

    GIc computes an `n_clusters`-partition
    based on a pre-computed minimum spanning tree. Clusters are merged
    so as to maximise (heuristically) the information
    criterion discussed in [2]_.

    GIc uses a bottom-up, agglomerative approach (as opposed to the ITM,
    which follows a divisive scheme). It greedily selects for merging
    a pair of clusters that maximises the information criterion [2]_.
    By default, the initial partition is determined by considering
    the intersection of the partitions found by multiple runs of
    the Genie++ method with thresholds [0.1, 0.3, 0.5, 0.7], which
    is a sensible choice for most clustering activities. Hence, contrary
    to the Genie method, we can say that GIc as virtually parameter-less.


    :Environment variables:
        OMP_NUM_THREADS
            See `genieclust.Genie`.


    References
    ----------

    .. [1]
        Cena A., *Adaptive hierarchical clustering algorithms based on
        data aggregation methods*, PhD Thesis, Systems Research Institute,
        Polish Academy of Sciences 2018.

    .. [2]
        Mueller A., Nowozin S., Lampert C.H., Information Theoretic
        Clustering using Minimum Spanning Trees, *DAGM-OAGM*, 2012.

    .. [3]
        Gagolewski M., Bartoszuk M., Cena A.,
        Genie: A new, fast, and outlier-resistant hierarchical clustering
        algorithm, *Information Sciences* 363, 2016, 8-23.
        doi:10.1016/j.ins.2016.05.003.

    """
    def __init__(
            self,
            n_clusters=2,
            *,
            gini_thresholds=[0.1, 0.3, 0.5, 0.7],
            M=1,
            affinity="l2",
            exact=True,
            compute_full_tree=False,
            compute_all_cuts=False,
            postprocess="boundary",
            cast_float32=True,
            mlpack_enabled="auto",
            mlpack_leaf_size=1,
            nmslib_n_neighbors=64,
            nmslib_params_init=dict(method="hnsw"),
            nmslib_params_index=dict(post=2),
            nmslib_params_query=dict(),
            add_clusters=0,
            n_features=None,
            verbose=False):
        # # # # # # # # # # # #
        super().__init__(
            n_clusters=n_clusters,
            M=M,
            affinity=affinity,
            exact=exact,
            compute_full_tree=compute_full_tree,
            compute_all_cuts=compute_all_cuts,
            postprocess=postprocess,
            cast_float32=cast_float32,
            mlpack_enabled=mlpack_enabled,
            mlpack_leaf_size=mlpack_leaf_size,
            nmslib_n_neighbors=nmslib_n_neighbors,
            nmslib_params_init=nmslib_params_init,
            nmslib_params_index=nmslib_params_index,
            nmslib_params_query=nmslib_params_query,
            verbose=verbose)

        self.gini_thresholds = gini_thresholds
        self.n_features      = n_features
        self.add_clusters    = add_clusters

        self._check_params()



    def _check_params(self, cur_state=None):
        cur_state = super()._check_params(cur_state)

        cur_state["add_clusters"] = int(self.add_clusters)
        if cur_state["add_clusters"] < 0:
            raise ValueError("`add_clusters` must be non-negative.")

        cur_state["gini_thresholds"] = np.array(self.gini_thresholds)
        for g in cur_state["gini_thresholds"]:
            if not (0.0 <= g <= 1.0):
                raise ValueError("All elements in `gini_thresholds` "
                                 "must be in [0,1].")

        return cur_state


    def fit(self, X, y=None):
        """
        Perform cluster analysis of a dataset.


        Parameters
        ----------

        X : object
            See `genieclust.Genie.fit`.

        y : None
            Ignored.


        Returns
        -------

        self : genieclust.GIc
            The object that the method was called on.


        See also
        --------

        genieclust.Genie.fit

        genieclust.GIc.fit_predict


        Notes
        -----

        Refer to the `labels_` and `n_clusters_` attributes for the result.

        Note that for `affinity` of ``"precomputed"``, the `n_features`
        parameter must be set explicitly.

        """
        cur_state = self._check_params()  # re-check, they might have changed

        cur_state = self._get_mst(X, cur_state)

        if cur_state["n_features"] < 1:  # _get_mst sets this
            # this shouldn't happen in normal use
            raise ValueError("Please set the `n_features` attribute manually.")

        if cur_state["verbose"]:
            print("[genieclust] Determining clusters with GIc.", file=sys.stderr)

        # apply the Genie+Ic algorithm:
        res = internal.gic_from_mst(
            self._mst_dist_,
            self._mst_ind_,
            n_features=cur_state["n_features"],
            n_clusters=cur_state["n_clusters"],
            add_clusters=cur_state["add_clusters"],
            gini_thresholds=cur_state["gini_thresholds"],
            noise_leaves=(cur_state["M"] > 1),
            compute_full_tree=cur_state["compute_full_tree"],
            compute_all_cuts=cur_state["compute_all_cuts"])

        cur_state = self._postprocess_outputs(res, cur_state)

        if cur_state["verbose"]:
            print("[genieclust] Done.", file=sys.stderr)

        return self



# distutils: language=c++
# cython: boundscheck=False
# cython: cdivision=True
# cython: nonecheck=False
# cython: wraparound=False
# cython: language_level=3



"""
Provides access to the Cargsort() and Cargkmin() functions.

Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License
Version 3, 19 November 2007, published by the Free Software Foundation.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License Version 3 for more details.
You should have received a copy of the License along with this program.
If not, see <https://www.gnu.org/licenses/>.
"""


cdef extern from "../src/c_argfuns.h":
    void Cargsort[T](ssize_t* ret, T* x, ssize_t n, bint stable)
    ssize_t Cargkmin[T](T* x, ssize_t n, ssize_t k, ssize_t* buf)



# distutils: language=c++
# cython: boundscheck=False
# cython: cdivision=True
# cython: nonecheck=False
# cython: wraparound=False
# cython: language_level=3



"""
Provides access to the CGenie and CGIc classes.

Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License
Version 3, 19 November 2007, published by the Free Software Foundation.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License Version 3 for more details.
You should have received a copy of the License along with this program.
If not, see <https://www.gnu.org/licenses/>.
"""


cdef extern from "../src/c_genie.h":
    cdef cppclass CGenie[T]:
        CGenie() except +
        CGenie(T* mst_d, ssize_t* mst_i, ssize_t n, bint noise_leaves, bint new_merge) except +
        void apply_genie(ssize_t n_clusters, double gini_threshold)  except +
        ssize_t get_max_n_clusters()
        ssize_t get_links(ssize_t* res)
        ssize_t get_labels(ssize_t n_clusters, ssize_t* res)
        void get_labels_matrix(ssize_t n_clusters, ssize_t* res)

    cdef cppclass CGIc[T]:
        CGIc() except +
        CGIc(T* mst_d, ssize_t* mst_i, ssize_t n, bint noise_leaves) except +
        void apply_gic(ssize_t n_clusters, ssize_t add_clusters,
            double n_features, double* gini_thresholds, ssize_t n_thresholds)  except +
        ssize_t get_max_n_clusters()
        ssize_t get_links(ssize_t* res)
        ssize_t get_labels(ssize_t n_clusters, ssize_t* res)
        void get_labels_matrix(ssize_t n_clusters, ssize_t* res)



"""
Various plotting functions
"""

# Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License
# Version 3, 19 November 2007, published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License Version 3 for more details.
# You should have received a copy of the License along with this program.
# If not, see <https://www.gnu.org/licenses/>.


import matplotlib.pyplot as plt
import matplotlib.colors
import numpy as np


# module globals:
col = ["k", "r", "g", "b", "c", "m", "y"]                                   + \
    [matplotlib.colors.to_hex(c) for c in plt.cm.get_cmap("tab10").colors]  + \
    [matplotlib.colors.to_hex(c) for c in plt.cm.get_cmap("tab20").colors]  + \
    [matplotlib.colors.to_hex(c) for c in plt.cm.get_cmap("tab20b").colors] + \
    [matplotlib.colors.to_hex(c) for c in plt.cm.get_cmap("tab20c").colors]

mrk = ["o", "^", "+", "x", "D", "v", "s", "*", "<", ">", "2"]



def _get_xy(X, y):
    # auxiliary function
    X = np.array(X)
    if X.ndim == 2:
        if not X.shape[1] == 2:
            raise ValueError("`X` must have 2 columns.")
    elif X.ndim == 1:
        if y is None:
            raise ValueError("If `X` is a vector, `y` should be provided.")

        y = np.array(y)
        if y.ndim != 1 or X.shape[0] != y.shape[0]:
            raise ValueError("`X` and `y` must have the same shape.")
        X = np.column_stack((X, y))
    else:
        raise ValueError("Invalid `X`.")

    return X


def plot_scatter(X, y=None, labels=None, **kwargs):
    """
    Draws a scatter plot


    Parameters
    ----------

    X : array_like
        Either a two-column matrix that gives the *x* and *y* coordinates of the points
        or a vector of length *n*. In the latter case, `y` must be a
        vector of length *n* as well.

    y : None or array_like
        The *y* coordinates of the *n* points in the case where `X` is a vector.

    labels : None or array_like
        A vector of *n* integer labels that correspond to each point in `X`,
        that gives its plot style.

    **kwargs : Collection properties
        Further arguments to `matplotlib.pyplot.scatter`.


    Notes
    -----

    If `X` is a two-column matrix, then ``plot_scatter(X)``
    is equivalent to ``plot_scatter(X[:,0], X[:,1])``.

    Unlike in `matplotlib.pyplot.scatter`,
    for any fixed ``j``, all points ``X[i,:]`` such that ``labels[i] == j``
    are always drawn in the same way, no matter the ``max(labels)``.
    In particular, labels 0, 1, 2, and 3 correspond to
    black, red, green, and blue, respectively.

    This function was inspired by the ``plot()`` function
    from the R package ``graphics``.


    Examples
    --------

    .. plot::

        An example scatter plots where each point is assigned one of
        two distinct labels:

        >>> n = np.r_[100, 50]
        >>> X = np.r_[np.random.randn(n[0], 2), np.random.randn(n[1], 2)+2.0]
        >>> l = np.repeat([0, 1], n)
        >>> genieclust.plots.plot_scatter(X, labels=l)
        >>> plt.show()                                   # doctest: +SKIP


    .. plot::

        Here are the first 10 plotting styles:

        >>> ncol = len(genieclust.plots.col)
        >>> nmrk = len(genieclust.plots.mrk)
        >>> mrk_recycled = np.tile(
        ...     genieclust.plots.mrk,
        ...     int(np.ceil(ncol/nmrk)))[:ncol]
        >>> for i in range(10):                               # doctest: +SKIP
        ...     plt.text(                                     # doctest: +SKIP
        ...         i, 0, i, horizontalalignment="center")    # doctest: +SKIP
        ...     plt.plot(                                     # doctest: +SKIP
        ...         i, 1, marker=mrk_recycled[i],             # doctest: +SKIP
        ...         color=genieclust.plots.col[i],            # doctest: +SKIP
        ...         markersize=25)                            # doctest: +SKIP
        >>> plt.title("Plotting styles for labels=0,1,...,9") # doctest: +SKIP
        >>> plt.ylim(-3,4)                                    # doctest: +SKIP
        >>> plt.axis("off")                                   # doctest: +SKIP
        >>> plt.show()                                        # doctest: +SKIP

    """
    X = _get_xy(X, y)

    if labels is None:
        labels = np.repeat(0, X.shape[0])
    else:
        labels = np.array(labels)
    if labels.ndim != 1 or X.shape[0] != labels.shape[0]:
        raise ValueError("Incorrect shape of `labels`.")

    for i in np.unique(labels):  # 0 is black, 1 is red, etc.
        plt.scatter(
            X[labels == i, 0],
            X[labels == i, 1],
            c=col[(i) % len(col)],
            marker=mrk[(i) % len(mrk)],
            **kwargs)



def plot_segments(pairs, X, y=None, style="k-", **kwargs):
    """
    Draws a set of disjoint line segments


    Parameters
    ----------

    pairs : array_like
        A two-column matrix that gives the pairs of indices
        defining the line segments to draw.

    X : array_like
        Either a two-column matrix that gives the *x* and *y* coordinates of the points
        or a vector of length *n*. In the latter case, `y` must be a
        vector of length *n* as well.

    y : None or array_like
        The *y* coordinates of the *n* points in the case where `X` is a vector.

    style:
        See `matplotlib.pyplot.plot`.

    **kwargs : Collection properties
        Further arguments to `matplotlib.pyplot.plot`.


    Notes
    -----

    The function draws a set of disjoint line segments from
    ``(X[pairs[i,0],0], X[pairs[i,0],1])`` to
    ``(X[pairs[i,1],0], X[pairs[i,1],1])``
    for all ``i`` from ``0`` to ``pairs.shape[0]-1``.

    `matplotlib.pyplot.plot` is called only once.
    Therefore, you can expect it to be pretty pretty fast.


    Examples
    --------

    .. plot::

        Plotting the convex hull of a point set:

        >>> import scipy.spatial
        >>> X = np.random.randn(100, 2)
        >>> hull = scipy.spatial.ConvexHull(X)
        >>> genieclust.plots.plot_scatter(X)
        >>> genieclust.plots.plot_segments(hull.simplices, X, style="r--")
        >>> plt.show()                                 # doctest: +SKIP


    .. plot::

        Plotting the minimum spanning tree:

        >>> X = np.random.randn(100, 2)
        >>> mst = genieclust.internal.mst_from_distance(X, "euclidean")
        >>> genieclust.plots.plot_scatter(X)
        >>> genieclust.plots.plot_segments(mst[1], X, style="m-.")
        >>> plt.axis("equal")                          # doctest: +SKIP
        >>> plt.show()                                 # doctest: +SKIP



    """
    X = _get_xy(X, y)

    pairs = np.array(pairs)
    if pairs.ndim != 2 or pairs.shape[1] != 2:
        raise ValueError("`pairs` must be a matrix with 2 columns.")

    xcoords = np.insert(X[pairs.ravel(), 0].reshape(-1, 2), 2, None, 1).ravel()
    ycoords = np.insert(X[pairs.ravel(), 1].reshape(-1, 2), 2, None, 1).ravel()
    plt.plot(xcoords, ycoords, style, **kwargs)



"""
genieclust Python Package

Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License
Version 3, 19 November 2007, published by the Free Software Foundation.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License Version 3 for more details.
You should have received a copy of the License along with this program.
If not, see <https://www.gnu.org/licenses/>.
"""

from . import plots
from . import inequity
from . import tools
from . import compare_partitions
from . import internal
from .genie import Genie, GIc

__version__ = '0.9.1' # see also ../setup.py



# distutils: language=c++
# cython: boundscheck=False
# cython: cdivision=True
# cython: nonecheck=False
# cython: wraparound=False
# cython: language_level=3



"""
Provides access to MST-related functions.

Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License
Version 3, 19 November 2007, published by the Free Software Foundation.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License Version 3 for more details.
You should have received a copy of the License along with this program.
If not, see <https://www.gnu.org/licenses/>.
"""


cdef extern from "../src/c_mst.h":

    cdef cppclass CDistance[T]:
        pass

    cdef cppclass CDistanceMutualReachability[T]: # inherits from CDistance
        CDistanceMutualReachability()
        CDistanceMutualReachability(const T* d_core, ssize_t n, CDistance[T]* d_pairwise)

    cdef cppclass CDistanceEuclidean[T]: # inherits from CDistance
        CDistanceEuclidean()
        CDistanceEuclidean(T* X, ssize_t n, ssize_t d)

    cdef cppclass CDistanceEuclideanSquared[T]: # inherits from CDistance
        CDistanceEuclideanSquared()
        CDistanceEuclideanSquared(T* X, ssize_t n, ssize_t d)

    cdef cppclass CDistanceManhattan[T]: # inherits from CDistance
        CDistanceManhattan()
        CDistanceManhattan(T* X, ssize_t n, ssize_t d)

    cdef cppclass CDistanceCosine[T]: # inherits from CDistance
        CDistanceCosine()
        CDistanceCosine(T* X, ssize_t n, ssize_t d)

    cdef cppclass CDistancePrecomputedMatrix[T]: # inherits from CDistance
        CDistancePrecomputedMatrix()
        CDistancePrecomputedMatrix(T* d, ssize_t n)

    cdef cppclass CDistancePrecomputedVector[T]: # inherits from CDistance
        CDistancePrecomputedVector()
        CDistancePrecomputedVector(T* d, ssize_t n)



    # cdef cppclass CMstTriple[T]:
    #     CMstTriple(ssize_t i1, ssize_t i2, T d, bint order=False)

    # ssize_t Cmst_from_nn_list[T](
    #     CMstTriple[T]* nns, ssize_t c,
    #     ssize_t n, T* mst_dist, ssize_t* mst_ind, bint verbose) except +


    ssize_t Cmst_from_nn[T](
        T* dist, ssize_t* ind, const T* d_core, ssize_t n, ssize_t k,
        T* mst_dist, ssize_t* mst_ind, bint* maybe_inexact, bint verbose) except +

    void Cknn_from_complete[T](
        CDistance[T]* D, ssize_t n, ssize_t k,
        T* dist, ssize_t* ind, bint verbose) except +

    void Cmst_from_complete[T](
        CDistance[T]* D, ssize_t n,
        T* mst_dist, ssize_t* mst_ind, bint verbose) except +



# distutils: language=c++
# cython: boundscheck=False
# cython: cdivision=True
# cython: nonecheck=False
# cython: wraparound=False
# cython: language_level=3


"""
Inequity (inequality) measures
"""


# Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License
# Version 3, 19 November 2007, published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License Version 3 for more details.
# You should have received a copy of the License along with this program.
# If not, see <https://www.gnu.org/licenses/>.


cimport cython
cimport numpy as np
import numpy as np
from . cimport c_inequity

ctypedef fused T:
    int
    long
    long long
    ssize_t
    float
    double

cdef T square(T x):
    return x*x



cpdef double gini_index(np.ndarray[T] x, bint is_sorted=False):
    """
    genieclust.inequity.gini_index(x, is_sorted=False)

    Computes the normalised Gini index


    Parameters
    ----------

    x : ndarray
        A vector with non-negative elements.

    is_sorted : bool
        Indicates if `x` is already sorted increasingly.


    Returns
    -------

    double
        The value of the inequity index, a number in [0,1].


    See Also
    --------

    genieclust.inequity.bonferroni_index : The normalised Bonferroni index


    Notes
    -----

    The normalised Gini [1]_ index is given by:

    .. math::

        G(x_1,\\dots,x_n) = \\frac{
        \\sum_{i=1}^{n-1} \\sum_{j=i+1}^n |x_i-x_j|
        }{
        (n-1) \\sum_{i=1}^n x_i
        }.

    Time complexity is :math:`O(n)` for sorted data; it holds:

    .. math::

        G(x_1,\\dots,x_n) = \\frac{
        \\sum_{i=1}^{n} (n-2i+1) x_{\\sigma(n-i+1)}
        }{
        (n-1) \\sum_{i=1}^n x_i
        },

    where :math:`\\sigma` is an ordering permutation of :math:`(x_1,\\dots,x_n)`.


    Both the Gini and Bonferroni indices can be used to quantify the "inequity"
    of a numeric sample. They can be perceived as measures of data dispersion.
    For constant vectors (perfect equity), the indices yield values of 0.
    Vectors with all elements but one equal to 0 (perfect inequity),
    are assigned scores of 1.
    Both indices follow the Pigou-Dalton principle (are Schur-convex):
    setting :math:`x_i = x_i - h` and :math:`x_j = x_j + h` with :math:`h > 0`
    and :math:`x_i - h \\geq  x_j + h` (taking from the "rich" and giving away
    to the "poor") decreases the inequity.

    These indices have applications in economics, amongst others.
    The `Genie` clustering algorithm uses the Gini index as a measure
    of the inequality of cluster sizes.



    References
    ----------

    .. [1]
        Gini C., *Variabilita e Mutabilita*,
        Tipografia di Paolo Cuppini, Bologna, 1912.


    Examples
    --------

    No inequality (perfect equality):

    >>> round(genieclust.inequity.gini_index(np.r_[2, 2,  2, 2, 2]), 2)
    0.0

    One has it all (total inequity):

    >>> round(genieclust.inequity.gini_index(np.r_[0, 0, 10, 0, 0]), 2)
    1.0

    Give to the poor, take away from the rich:

    >>> round(genieclust.inequity.gini_index(np.r_[7, 0,  3, 0, 0]), 2)
    0.85

    Robinhood even more:

    >>> round(genieclust.inequity.gini_index(np.r_[6, 0,  3, 1, 0]), 2)
    0.75
    """

    if not is_sorted: x = np.sort(x)
    else: x = np.array(x, dtype=x.dtype, copy=False, order="C") # assure c_contiguity
    return c_inequity.Cgini_sorted(&x[0], x.shape[0])



cpdef double bonferroni_index(np.ndarray[T] x, bint is_sorted=False):
    """
    genieclust.inequity.bonferroni_index(x, is_sorted=False)

    Computes the normalised Bonferroni index


    Parameters
    ----------

    x : ndarray
        A vector with non-negative elements.

    is_sorted : bool
        Indicates if `x` is already sorted increasingly.


    Returns
    -------

    double
        The value of the inequity index, a number in [0,1].


    See Also
    --------

    genieclust.inequity.gini_index : The normalised Gini index


    Notes
    -----

    The normalised Bonferroni [1]_ index is given by:

    .. math::

        B(x_1,\\dots,x_n) = \\frac{
        \\sum_{i=1}^{n}  \\left( n-\\sum_{j=1}^i \\frac{n}{n-j+1} \\right) x_{\\sigma(n-i+1)}
        }{
        (n-1) \\sum_{i=1}^n x_i
        },

    where :math:`\\sigma` is an ordering permutation of :math:`(x_1,\\dots,x_n)`.

    Time complexity: :math:`O(n)` for sorted data.


    References
    ----------

    .. [1]
        Bonferroni C., *Elementi di Statistica Generale*, Libreria Seber,
        Firenze, 1930.


    Examples
    --------

    No inequality (perfect equality):

    >>> round(genieclust.inequity.bonferroni_index(np.r_[2, 2,  2, 2, 2]), 2)
    0.0

    One has it all (total inequity):

    >>> round(genieclust.inequity.bonferroni_index(np.r_[0, 0, 10, 0, 0]), 2)
    1.0

    Give to the poor, take away from the rich:

    >>> round(genieclust.inequity.bonferroni_index(np.r_[7, 0,  3, 0, 0]), 2)
    0.91

    Robinhood even more:

    >>> round(genieclust.inequity.bonferroni_index(np.r_[6, 0,  3, 1, 0]), 2)
    0.83
    """

    if not is_sorted: x = np.sort(x)
    else: x = np.array(x, dtype=x.dtype, copy=False, order="C") # assure c_contiguity

    return c_inequity.Cbonferroni_sorted(&x[0], x.shape[0])



# distutils: language=c++
# cython: boundscheck=False
# cython: cdivision=True
# cython: nonecheck=False
# cython: wraparound=False
# cython: language_level=3



## TODO: (good first issue) Implement argkmax


"""
Functions a person might find useful, but not necessarily
"""


# Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License
# Version 3, 19 November 2007, published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License Version 3 for more details.
# You should have received a copy of the License along with this program.
# If not, see <https://www.gnu.org/licenses/>.






cimport cython
cimport numpy as np
import numpy as np


cimport libc.math
from libcpp cimport bool
from libcpp.vector cimport vector



ctypedef fused T:
    int
    long
    long long
    ssize_t
    float
    double

ctypedef fused floatT:
    float
    double



from . cimport c_argfuns


################################################################################
# cummin and cummax
################################################################################


cpdef np.ndarray[ssize_t] cummin(T[:] x):
    """
    genieclust.tools.cummin(x)

    Cumulative minimum


    Parameters
    ----------

    x : ndarray
        A vector with *n* elements of type ``int``, ``float``,
        or ``double``.


    Returns
    -------

    ndarray
        A vector of length *n* whose ``i``-th element
        is equal to ``min(x[:i])``.



    See Also
    --------

    genieclust.tools.cummax : Cumulative maximum

    numpy.cumsum : Cumulative sum

    numpy.cumprod : Cumulative product


    Examples
    --------

    >>> genieclust.tools.cummin(np.r_[3, 4, 2, 1, 5, 6])
    array([3, 3, 2, 1, 1, 1])

    """
    cdef ssize_t n = x.shape[0], i
    cdef np.ndarray[T] ret = np.empty_like(x)
    ret[0] = x[0]
    for i in range(1, n):
        if x[i] <= ret[i-1]:
            ret[i] = x[i]
        else:
            ret[i] = ret[i-1]

    return ret


cpdef np.ndarray[ssize_t] cummax(T[:] x):
    """
    genieclust.tools.cummax(x)

    Cumulative maximum


    Parameters
    ----------

    x : ndarray
        A vector with *n* elements of type ``int``, ``float``,
        or ``double``.


    Returns
    -------

    ndarray
        A vector of length *n* whose ``i``-th element
        is equal to ``max(x[:i])``.



    See Also
    --------

    genieclust.tools.cummin : Cumulative minimum

    numpy.cumsum : Cumulative sum

    numpy.cumprod : Cumulative product


    Examples
    --------

    >>> genieclust.tools.cummax(np.r_[3, 2, 1, 4, 6, 5])
    array([3, 3, 3, 4, 6, 6])

    """
    cdef ssize_t n = x.shape[0], i
    cdef np.ndarray[T] ret = np.empty_like(x)
    ret[0] = x[0]
    for i in range(1, n):
        if x[i] >= ret[i-1]:
            ret[i] = x[i]
        else:
            ret[i] = ret[i-1]

    return ret





cpdef ssize_t argkmin(np.ndarray[T] x, int k):
    """
    genieclust.tools.argkmin(x, k)

    Finds the position of an order statistic


    Parameters
    ----------

    x : ndarray
        A vector with *n* elements of type ``int``, ``float``,
        or ``double``.

    k : int
        An integer between 0 and *n* - 1, preferably small.


    Returns
    -------

    int
        The index where the (`k`-1)-th smallest value in `x` is located.


    Notes
    -----

    It holds ``argkmin(x, 0) == argmin(x)``, or, more generally,
    ``argkmin(x, k) == np.argsort(x)[k]``.

    Run time is :math:`O(nk)` and working memory is :math:`O(k)`.
    An insertion sort-like scheme is used to locate the order statistic.
    In practice, the function is very fast for small `k` and randomly ordered
    or almost sorted (increasingly) data.

    ================================== =============== ==================
    Example timings                    `argkmin(x, k)` `np.argsort(x)[k]`
    ================================== =============== ==================
    (ascending)  n= 100000000, k=   1:      0.060s       4.388s
    (descending)                            0.168s       7.329s
    (random)                                0.073s      26.673s
    (ascending)  n= 100000000, k=   5:      0.060s       4.403s
    (descending)                            0.505s       7.414s
    (random)                                0.072s      26.447s
    (ascending)  n= 100000000, k= 100:      0.061s       4.390s
    (descending)                            8.007s       7.639s
    (random)                                0.075s      27.299s
    ================================== =============== ==================


    Examples
    --------

    >>> x = np.r_[2, 3, 6, 5, 1, 4]
    >>> genieclust.tools.argkmin(x, 0) # index of the smallest value
    4
    >>> genieclust.tools.argkmin(x, 1) # index of the 2nd smallest value
    0
    """
    x = np.array(x, dtype=x.dtype, copy=False, order="C") # assure c_contiguity
    cdef ssize_t n = x.shape[0]
    return c_argfuns.Cargkmin(&x[0], n, k, <ssize_t*>0)



cpdef np.ndarray[ssize_t] _argsort(np.ndarray[T] x, bint stable=True):
    """
    genieclust.tools.argsort(x, stable=True)

    Finds the/an ordering permutation (provided for testing only)


    Parameters
    ----------

    x : ndarray
        A vector with *n* elements of type ``int``, ``float``,
        or ``double``.

    stable : bool
        Should a stable (a bit slower) sorting algorithm be used?


    Returns
    -------

    ndarray
        An ordering permutation of `x`, an integer vector `o` of length *n*
        with elements between 0 and *n* - 1
        such that ``x[o[0]] <= x[o[1]] <= ... <= x[o[n-1]]``.


    Notes
    -----

    Finds the/an ordering permutation of a ``c_contiguous`` array `x`
    The ordering permutation is uniquely defined provided that
    `stable` is ``True``. Otherwise *an* ordering permutation
    will be generated.

    """
    x = np.array(x, dtype=x.dtype, copy=False, order="C") # assure c_contiguity
    cdef ssize_t n = x.shape[0]
    cdef np.ndarray[ssize_t] ret = np.empty(n, dtype=np.intp)
    c_argfuns.Cargsort(&ret[0], &x[0], n, stable)
    return ret







cpdef np.ndarray[floatT] _core_distance(np.ndarray[floatT,ndim=2] dist, int M):
    """
    (provided for testing only)


    Given a pairwise distance matrix, computes the "core distance", i.e.,
    the distance of each point to its M-th nearest neighbour.
    Note that M==1 always yields all the distances equal to 0.0.
    The core distances are needed when computing the mutual reachability
    distance in the HDBSCAN* algorithm.

    See R. Campello, D. Moulavi, A. Zimek, J. Sander, Hierarchical density
    estimates for data clustering, visualization, and outlier detection,
    ACM Transactions on Knowledge Discovery from Data 10(1):5:1–5:51, 2015.
    doi: 10.1145/2733381.

    The input distance matrix for a given point cloud X may be computed,
    e.g., via a call to
    `scipy.spatial.distance.squareform(scipy.spatial.distance.pdist(X))`.


    Parameters
    ----------

    dist : ndarray, shape (n_samples,n_samples)
        A pairwise n*n distance matrix.
    M : int
        A smoothing factor >= 1.


    Returns
    -------

    d_core : ndarray, shape (n_samples,)
        d_core[i] gives the distance between the i-th point and its M-th nearest
        neighbour. The i-th point's 1st nearest neighbour is the i-th point itself.
    """
    cdef ssize_t n = dist.shape[0], i, j
    cdef floatT v
    cdef np.ndarray[floatT] d_core = np.zeros(n,
        dtype=np.float32 if floatT is float else np.float64)
    cdef floatT[::1] row

    if M < 1: raise ValueError("M < 1")
    if dist.shape[1] != n: raise ValueError("not a square matrix")
    if M >= n: raise ValueError("M >= matrix size")

    if M == 1: return d_core # zeros

    cdef vector[ssize_t] buf = vector[ssize_t](M)
    for i in range(n):
        row = dist[i,:]
        j = c_argfuns.Cargkmin(&row[0], row.shape[0], M-1, buf.data())
        d_core[i] = dist[i, j]

    return d_core


cpdef np.ndarray[floatT,ndim=2] _mutual_reachability_distance(
        np.ndarray[floatT,ndim=2] dist,
        np.ndarray[floatT] d_core):
    """
    (provided for testing only)


    Given a pairwise distance matrix,
    computes the mutual reachability distance w.r.t. the given
    core distance vector, see genieclust.internal.core_distance().

    See R. Campello, D. Moulavi, A. Zimek, J. Sander, Hierarchical density
    estimates for data clustering, visualization, and outlier detection,
    ACM Transactions on Knowledge Discovery from Data 10(1):5:1–5:51, 2015.
    doi: 10.1145/2733381.

    The input distance matrix for a given point cloud X
    may be computed, e.g., via a call to
    ``scipy.spatial.distance.squareform(scipy.spatial.distance.pdist(X))``.


    Parameters
    ----------

    dist : ndarray, shape (n_samples,n_samples)
        A pairwise n*n distance matrix.
    d_core : ndarray, shape (n_samples,)
        See genieclust.internal.core_distance().


    Returns
    -------

    R : ndarray, shape (n_samples,n_samples)
        A new distance matrix, giving the mutual reachability distance.
    """
    cdef ssize_t n = dist.shape[0], i, j
    cdef floatT v
    if dist.shape[1] != n: raise ValueError("not a square matrix")

    cdef np.ndarray[floatT,ndim=2] R = np.array(dist,
        dtype=np.float32 if floatT is float else np.float64)
    for i in range(0, n-1):
        for j in range(i+1, n):
            v = dist[i, j]
            if v < d_core[i]: v = d_core[i]
            if v < d_core[j]: v = d_core[j]
            R[i, j] = R[j, i] = v

    return R






"""
genieclust Package
Copyright (C) 2018-2020 Marek Gagolewski (https://www.gagolewski.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License
Version 3, 19 November 2007, published by the Free Software Foundation.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License Version 3 for more details.
You should have received a copy of the License along with this program.
If not, see <https://www.gnu.org/licenses/>.
"""

import setuptools
from distutils.extension import Extension
from distutils.command.sdist import sdist
from Cython.Distutils import build_ext
from Cython.Build import cythonize
import numpy as np
import os.path
import glob
import os
import sys


cython_modules = {
    "genieclust.internal": [
        os.path.join("genieclust", "internal.pyx")
    ],
    "genieclust.compare_partitions": [
        os.path.join("genieclust", "compare_partitions.pyx")
    ],
    "genieclust.inequity": [
        os.path.join("genieclust", "inequity.pyx")
    ],
    "genieclust.tools": [
        os.path.join("genieclust", "tools.pyx")
    ]
}


class genieclust_sdist(sdist):
    def run(self):
        for pyx_files in cython_modules.values():
            cythonize(pyx_files)
        sdist.run(self)


class genieclust_build_ext(build_ext):
    def build_extensions(self):

        # This code is adapted from
        # scikit-learn/sklearn/_build_utils/openmp_helpers.py
        # (version last updated on 13 Nov 2019; 9876f74)
        # See https://github.com/scikit-learn/scikit-learn and https://scikit-learn.org/.

        if hasattr(self.compiler, 'compiler'):
            compiler = self.compiler.compiler[0]
        else:
            compiler = self.compiler.__class__.__name__

        if sys.platform == "win32" and ('icc' in compiler or 'icl' in compiler):
            for e in self.extensions:
                e.extra_compile_args += ['/Qopenmp']
                e.extra_link_args += ['/Qopenmp']
        elif sys.platform == "win32":
            for e in self.extensions:
                e.extra_compile_args += ['/openmp']
                e.extra_link_args += ['/openmp']
        elif sys.platform == "darwin" and ('icc' in compiler or 'icl' in compiler):
            for e in self.extensions:
                e.extra_compile_args += ['-openmp']
                e.extra_link_args += ['-openmp']
        elif sys.platform == "darwin" and 'openmp' in os.getenv('CPPFLAGS', ''):
            # -fopenmp can't be passed as compile flag when using Apple-clang.
            # OpenMP support has to be enabled during preprocessing.
            #
            # For example, our macOS wheel build jobs use the following environment
            # variables to build with Apple-clang and the brew installed "libomp":
            #
            # export CPPFLAGS="$CPPFLAGS -Xpreprocessor -fopenmp"
            # export CFLAGS="$CFLAGS -I/usr/local/opt/libomp/include"
            # export CXXFLAGS="$CXXFLAGS -I/usr/local/opt/libomp/include"
            # export LDFLAGS="$LDFLAGS -Wl,-rpath,/usr/local/opt/libomp/lib
            #                          -L/usr/local/opt/libomp/lib -lomp"
            pass
        else:
            # Default flag for GCC and clang:
            for e in self.extensions:
                e.extra_compile_args += ['-fopenmp']
                e.extra_link_args += ['-fopenmp']

        # Old version:
        # c = self.compiler.compiler_type
        # if c == "msvc":
        #     for e in self.extensions:
        #         e.extra_compile_args += "/openmp"
        # elif c == "mingw32":
        #     for e in self.extensions:
        #         e.extra_compile_args += "-fopenmp"
        #         e.extra_link_args += "-fopenmp"
        # elif c == "unix":
        #     # Well... gcc/clang has -fopenmp,
        #     # icc has -openmp, oracle has -xopenmp, etc.
        #     # The user should specify CXXFLAGS and LDFLAGS herself, I think.
        #     pass

        build_ext.build_extensions(self)


ext_kwargs = dict(
    include_dirs=[np.get_include()],
    language="c++",
    depends=glob.glob(os.path.join("src", "c_*.h")) +
            glob.glob(os.path.join("genieclust", "*.pxd")),
    #define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
)


with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="genieclust",
    version="0.9.1", # see also genieclust/__init__.py
    description="The Genie++ Hierarchical Clustering Algorithm (with Extras)",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Marek Gagolewski",
    author_email="marek@gagolewski.com",
    maintainer="Marek Gagolewski",
    license="GNU Affero General Public License v3",
    install_requires=[
        "numpy",
        "scipy",
        "cython",
        "matplotlib",
        "sklearn"
      ],
    extras_require={
        "mlpack": ["mlpack"],
        "nmslib": ["nmslib"]
    },
    download_url="https://github.com/gagolews/genieclust",
    url="https://genieclust.gagolewski.com/",
    project_urls={
        "Bug Tracker":   "https://github.com/gagolews/genieclust/issues",
        "Documentation": "https://genieclust.gagolewski.com/",
        "Source Code":   "https://github.com/gagolews/genieclust",
    },
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering",
    ],
    cmdclass={
        "sdist": genieclust_sdist,
        "build_ext": genieclust_build_ext
    },
    ext_modules=[
        Extension(module, pyx_files, **ext_kwargs)
        for module, pyx_files in cython_modules.items()
    ]
)
